"""
Collection of NMEA helper methods which can be used
outside the NMEAMessage or NMEAReader classes

Created on 04 Mar 2021

:author: semuadmin (Steve Smith)
:copyright: semuadmin © 2021
:license: BSD 3-Clause
"""

# pylint: disable=invalid-name

import re
from datetime import datetime
from math import acos, asin, atan2, cos, pi, sin, sqrt

import pynmeagps.exceptions as nme
from pynmeagps.nmeatypes_core import (
    DM,
    DT,
    DTL,
    GPSEPOCH0,
    LA,
    LN,
    NMEA_MSGIDS,
    NMEA_MSGIDS_PROP,
    WGS84,
    WGS84_FLATTENING,
    WGS84_SMAJ_AXIS,
)

KNOTSCONV = {"MS": 0.5144447324, "FS": 1.68781084, "MPH": 1.15078, "KMPH": 1.852001}
LEAPS0 = datetime(1900, 1, 1, 0, 0, 0)
LEAPSECONDS = [
    (2272060800, 10),  # 1 Jan 1972
    (2287785600, 11),  # 1 Jul 1972
    (2303683200, 12),  # 1 Jan 1973
    (2335219200, 13),  # 1 Jan 1974
    (2366755200, 14),  # 1 Jan 1975
    (2398291200, 15),  # 1 Jan 1976
    (2429913600, 16),  # 1 Jan 1977
    (2461449600, 17),  # 1 Jan 1978
    (2492985600, 18),  # 1 Jan 1979
    (2524521600, 19),  # 1 Jan 1980
    (2571782400, 20),  # 1 Jul 1981
    (2603318400, 21),  # 1 Jul 1982
    (2634854400, 22),  # 1 Jul 1983
    (2698012800, 23),  # 1 Jul 1985
    (2776982400, 24),  # 1 Jan 1988
    (2840140800, 25),  # 1 Jan 1990
    (2871676800, 26),  # 1 Jan 1991
    (2918937600, 27),  # 1 Jul 1992
    (2950473600, 28),  # 1 Jul 1993
    (2982009600, 29),  # 1 Jul 1994
    (3029443200, 30),  # 1 Jan 1996
    (3076704000, 31),  # 1 Jul 1997
    (3124137600, 32),  # 1 Jan 1999
    (3345062400, 33),  # 1 Jan 2006
    (3439756800, 34),  # 1 Jan 2009
    (3550089600, 35),  # 1 Jul 2012
    (3644697600, 36),  # 1 Jul 2015
    (3692217600, 37),  # 1 Jan 2017
]
"""
Leapsecond list from NTC - Updated through IERS Bulletin C69
https://hpiers.obspm.fr/iers/bul/bulc/ntp/leap-seconds.list
File expires on: 28 June 2026
"""


def area(
    lat1: float,
    lon1: float,
    lat2: float,
    lon2: float,
    radius: int = WGS84_SMAJ_AXIS / 1000,
) -> float:
    """
    Calculate spherical area bounded by two coordinates.

    :param float lat1: lat1
    :param float lon1: lon1
    :param float lat2: lat2
    :param float lon2: lon2
    :param float radius: radius in km (Earth = 6378.137 km)
    :return: area in km²
    :rtype: float
    """

    phi1, phi2 = [c * pi / 180 for c in (lat1, lat2)]
    return pow(radius, 2) * pi * abs(sin(phi1) - sin(phi2)) * abs(lon1 - lon2) / 180


def bearing(lat1: float, lon1: float, lat2: float, lon2: float) -> float:
    """
    Calculate bearing between two coordinates.

    :param float lat1: lat1
    :param float lon1: lon1
    :param float lat2: lat2
    :param float lon2: lon2
    :return: bearing in degrees
    :rtype: float
    """

    phi1, lambda1, phi2, lambda2 = [c * pi / 180 for c in (lat1, lon1, lat2, lon2)]
    y = sin(lambda2 - lambda1) * cos(phi2)
    x = cos(phi1) * sin(phi2) - sin(phi1) * cos(phi2) * cos(lambda2 - lambda1)
    theta = atan2(y, x)
    brng = (theta * 180 / pi + 360) % 360

    return brng


def calc_checksum(content: object) -> str:
    """
    Calculate checksum for raw NMEA message.

    :param object content: NMEA message content (everything except checksum)
    :return: checksum as hex string
    :rtype: str
    """

    cksum = 0
    for sub in content:
        cksum ^= ord(sub)
    return f"{cksum:02X}"


def date2str(dat: datetime.date, form: str = DT) -> str:
    """
    Convert datetime.date to NMEA formatted string.

    :param datetime.date dat: date
    :param str form: date format DT = ddmmyy, DTL = ddmmyyyy, DM = mmddyy (DT)
    :return: NMEA formatted date string
    :rtype: str
    """

    try:
        if form == DM:
            dform = "%m%d%y"
        elif form == DTL:
            dform = "%d%m%Y"
        else:
            dform = "%d%m%y"
        return dat.strftime(dform)
    except (AttributeError, TypeError, ValueError):
        return ""


def date2utc(dates: str, form: str = DT) -> datetime.date:
    """
    Convert NMEA Date to UTC datetime.

    :param str dates: NMEA date
    :param str form: date format DT = ddmmyy, DTL = ddmmyyyy, DM = mmddyy (DT)
    :return: UTC date YYyy:mm:dd
    :rtype: datetime.date
    """

    try:
        if form == "DM":
            dform = "%m%d%y"
        elif form == "DTL":
            dform = "%d%m%Y"
        else:
            dform = "%d%m%y"
        utc = datetime.strptime(dates, dform)
        return utc.date()
    except (TypeError, ValueError):
        return ""


def ddd2dmm(degrees: float, att: str, hpmode: bool = False) -> str:
    """
    Convert decimal degrees to native NMEA degrees decimal
    minutes string (NB: standard NMEA only supports 5dp
    minutes precision - a high precision mode offers 7dp
    precision but this may not be accepted by all NMEA parsers).

    :param float degrees: degrees
    :param str att: 'LA' (lat) or 'LN' (lon)
    :param bool hpmode: high precision mode (7dp rather than 5dp)
    :return: degrees as (d)ddmm.mmmmm(mm) formatted string
    :rtype: str

    """

    try:
        degrees = abs(degrees)
        degrees, minutes = divmod(degrees * 60, 60)
        degrees = int(degrees * 100)
        if hpmode:
            if att == LA:
                dmm = f"{degrees + minutes:.7f}".zfill(12)
            else:  # LN
                dmm = f"{degrees + minutes:.7f}".zfill(13)
        else:
            if att == LA:
                dmm = f"{degrees + minutes:.5f}".zfill(10)
            else:  # LN
                dmm = f"{degrees + minutes:.5f}".zfill(11)
        return dmm
    except (TypeError, ValueError):
        return ""


def deg2dmm(degrees: float, att: str) -> str:
    """
    Convert decimal degrees to degrees decimal minutes string
    e.g. '51°20.76′S'.

    :param float degrees: degrees
    :param str att: 'LA' (lat) or 'LN' (lon)
    :return: degrees as dm.m formatted string
    :rtype: str

    """

    try:
        negative = degrees < 0
        degrees = abs(degrees)
        degrees, minutes = divmod(degrees * 60, 60)
        if negative:
            sfx = "S" if att == LA else "W"
        else:
            sfx = "N" if att == LA else "E"
        return f"{int(degrees)}\u00b0{round(minutes,7)}\u2032{sfx}"
    except (TypeError, ValueError):
        return ""


def deg2dms(degrees: float, att: str) -> str:
    """
    Convert decimal degrees to degrees minutes seconds string
    e.g. '51°20′45.6″N'

    :param float degrees: degrees
    :param str att: 'LA' (lat) or 'LN' (lon)
    :return: degrees as d.m.s formatted string
    :rtype: str

    """

    try:
        negative = degrees < 0
        degrees = abs(degrees)
        minutes, seconds = divmod(degrees * 3600, 60)
        degrees, minutes = divmod(minutes, 60)
        if negative:
            sfx = "S" if att == LA else "W"
        else:
            sfx = "N" if att == LA else "E"
        return f"{int(degrees)}\u00b0{int(minutes)}\u2032{round(seconds,5)}\u2033{sfx}"
    except (TypeError, ValueError):
        return ""


def dmm2ddd(pos: str) -> float:
    """
    Convert NMEA lat/lon string to (unsigned) decimal degrees.

    :param str pos: (d)ddmm.mmmmm
    :return: pos as decimal degrees
    :rtype: float or str if invalid

    """

    try:
        dp = pos.find(".")
        if dp < 4:
            raise ValueError()
        posdeg = float(pos[0 : dp - 2])
        posmin = float(pos[dp - 2 :])
        return round((posdeg + posmin / 60), 10)
    except (TypeError, ValueError):
        return ""


def dms2deg(
    dmsstr: str,
    rnd: int = 7,
) -> float:
    """
    Convert degrees, minutes, seconds string to decimal degrees.

    Expects degrees and (optional) minutes and seconds to be delimited by
    non-numeric char, e.g.

    '51°20′45.6″N' -> 51.346

    '51°30.0′' -> 51.5

    :param str dmsstr: d,m,s string
    :param int rnd: decimal places (7)
    :return: degrees as d.dd
    :rtype: float
    :raises: ValueError if invalid dms string format
    """

    dmsstr = dmsstr.replace(" ", "")
    if (
        re.match(
            r"^([\d.]+[^\d.]{1}){1}([\d.]+[^\d.]{1})?([\d.]+[^\d.]{1})?[NSEW]?$", dmsstr
        )
        is None
    ):
        raise ValueError(f"Invalid dms string format {dmsstr}")

    sign = -1 if dmsstr.find("S") >= 0 or dmsstr.find("W") >= 0 else 1
    dms = re.split(r"[^\d.]", dmsstr, maxsplit=3)
    ddd = 0.0
    for x in range(len(dms) - 1):
        ddd += 0.0 if dms[x] == "" else float(dms[x]) / pow(60, x)
    return round(ddd * sign, rnd)


def ecef2llh(
    x: float,
    y: float,
    z: float,
    a: float = WGS84_SMAJ_AXIS,
    f: float = WGS84_FLATTENING,
) -> tuple:
    """
    Convert ECEF coordinates to geodetic (LLH) using Olson algorithm.

    Olson, D. K. (1996). Converting Earth-Centered, Earth-Fixed Coordinates to
    Geodetic Coordinates. IEEE Transactions on Aerospace and Electronic Systems,
    32(1), 473-476. https://doi.org/10.1109/7.481290

    :param float x: X coordinate
    :param float y: Y coordinate
    :param float z: Z coordinate
    :param float a: semi-major axis (6378137.0 for WGS84)
    :param float f: flattening (298.257223563 for WGS84)
    :return: tuple of (lat, lon, ellipsoidal height in m) as floats
    :rtype: tuple
    """
    # pylint: disable=too-many-locals

    # commented default values are for WGS84 spheroid
    f = 1 / f
    e2 = f * (2 - f)  # 6.6943799901377997e-3
    a1 = a * e2  # 4.2697672707157535e4
    a2 = a1 * a1  # 1.8230912546075455e9
    a3 = a1 * e2 / 2  # 1.8230912546075455e9
    a4 = 2.5 * a2  # 4.5577281365188637e9
    a5 = a1 + a3  # 4.2840589930055659e4
    a6 = 1 - e2  # 9.9330562000986220e-1
    zp = abs(z)
    w2 = x * x + y * y
    w = sqrt(w2)
    z2 = z * z
    r2 = w2 + z2
    r = sqrt(r2)

    # algorithm inaccurate near Earth's core
    # so nominal value returned
    if r < 100000.0:
        return 0.0, 0.0, -1.0e7

    lon = atan2(y, x)
    s2 = z2 / r2
    c2 = w2 / r2
    u = a2 / r
    v = a3 - a4 / r
    if c2 > 0.3:
        s = (zp / r) * (1.0 + c2 * (a1 + u + s2 * v) / r)
        lat = asin(s)
        ss = s * s
        c = sqrt(1.0 - ss)
    else:
        c = (w / r) * (1.0 - s2 * (a5 - u - c2 * v) / r)
        lat = acos(c)
        ss = 1.0 - c * c
        s = sqrt(ss)
    g = 1.0 - e2 * ss
    rg = a / sqrt(g)
    rf = a6 * rg
    u = w - rg * c
    v = zp - rf * s
    f = c * u + s * v
    m = c * v - s * u
    p = m / (rf / g + f)
    lat = lat + p
    height = f + m * p / 2.0
    if z < 0.0:
        lat = -lat

    lat, lon = [c * 180 / pi for c in (lat, lon)]
    return lat, lon, height


def generate_checksum(talker: str, msgID: str, payload: list) -> str:
    """
    Generate checksum for new NMEA message.

    :param str talker: talker e.g. "GN"
    :param str msgID: msgID e.g. "GLL"
    :param list payload: payload as list
    :return: checksum as hex string
    :rtype: str
    """

    content = talker + msgID
    for field in payload:
        content += f",{field}"
    return calc_checksum(content)


def get_gpswnotow(dat: datetime) -> tuple:
    """
    Get GPS Week number (Wno) and Time of Week (Tow)
    for midnight on given date.

    GPS Epoch 0 = 6th Jan 1980

    :param datetime dat: calendar date
    :return: tuple of (Wno, Tow)
    :rtype: tuple
    """

    wno = int((dat - GPSEPOCH0).days / 7)
    tow = ((dat.weekday() + 1) % 7) * 86400
    return wno, tow


def get_parts(message: object) -> tuple:
    """
    Get content, talker, msgid, payload and checksum of raw NMEA message.

    :param object message: entire message as bytes or string
    :return: tuple of (content, talker, msgID, payload as list, checksum)
    :rtype: tuple
    :raises: NMEAMessageError (if message is badly formed)
    """

    try:
        if isinstance(message, bytes):
            message = message.decode("utf-8")
        content, cksum = message.strip("$\r\n").split("*", 1)
        hdr, *payload = content.split(",")
        s = 1 if hdr[:1] == "P" else 2
        talker, msgid = hdr[:s], hdr[s:]
        return content, talker, msgid, payload, cksum
    except Exception as err:
        raise nme.NMEAMessageError(f"Badly formed message {message}") from err


def haversine(
    lat1: float,
    lon1: float,
    lat2: float,
    lon2: float,
    radius: int = WGS84_SMAJ_AXIS / 1000,
) -> float:
    """
    Calculate spherical distance in km between two coordinates using haversine formula.

    NB suitable for coordinates greater than around 50m apart. For
    smaller separations, use the planar approximation formula.

    :param float lat1: lat1
    :param float lon1: lon1
    :param float lat2: lat2
    :param float lon2: lon2
    :param float radius: radius in km (Earth = 6378.137 km)
    :return: spherical distance in km
    :rtype: float
    """

    phi1, lambda1, phi2, lambda2 = [c * pi / 180 for c in (lat1, lon1, lat2, lon2)]
    dist = radius * acos(
        cos(phi2 - phi1) - cos(phi1) * cos(phi2) * (1 - cos(lambda2 - lambda1))
    )

    return dist


def hex2str(num: int, padding: int = 0) -> str:
    """
    Convert hex integer to padded or unpadded string format,
    as used by some proprietary NMEA message types.

    :param int num: hexadecimal integer
    :param padding: no of padded digits (0)
    :return: integer as string
    :rtype: str
    """

    pad = f"{padding:02d}" if padding else ""
    return f"{num:{pad}x}".upper()


def knots2spd(knots: float, unit: str = "MS") -> float:
    """
    Convert speed in knots to speed in specified units.

    :param float knots: knots
    :param unit str: 'MS' (default), 'FS', MPH', 'KMPH'
    :return: speed in m/s, feet/s, mph or kmph
    :rtype: float

    """

    try:
        return knots * KNOTSCONV[unit.upper()]
    except KeyError as err:
        raise KeyError(
            f"Invalid conversion unit {unit.upper()} - must be in {list(KNOTSCONV.keys())}."
        ) from err
    except TypeError as err:
        raise TypeError(
            f"Invalid knots value {knots} - must be float or integer."
        ) from err


def latlon2dmm(lat: float, lon: float) -> tuple:
    """
    Converts decimal lat/lon tuple to degrees decimal minutes.

    :param float lat: lat
    :param float lon: lon
    :return: (lat,lon) in d.mm.m format
    :rtype: tuple
    """

    lat = deg2dmm(lat, LA)
    lon = deg2dmm(lon, LN)
    return lat, lon


def latlon2dms(lat: float, lon: float) -> tuple:
    """
    Converts decimal lat/lon tuple to degrees minutes seconds.

    :param float lat: lat
    :param float lon: lon
    :return: (lat,lon) in d.m.s. format
    :rtype: tuple
    """

    lat = deg2dms(lat, LA)
    lon = deg2dms(lon, LN)
    return lat, lon


def leapsecond(dat: datetime) -> int:
    """
    Get GPS UTC leapsecond offset effective at date.

    Refer to LEAPSECONDS constant.

    - LEAPS0 = 1900-01-01 00:00:00
    - GPSEPOCH0 = 1980-01-06: 00:00:00

    :param datetime.datetime dat: effective date
    :return: leapsecond offset
    :rtype: int
    """

    def totsec(dt) -> int:
        return (dt - LEAPS0).total_seconds()

    if totsec(dat) < LEAPSECONDS[0][0]:
        return 0

    dts = []
    for dt in (dat, GPSEPOCH0):
        dts.append(max((val for key, val in LEAPSECONDS if key < totsec(dt))))
    return dts[0] - dts[1]


def llh2ecef(
    lat: float,
    lon: float,
    height: float,
    a: float = WGS84_SMAJ_AXIS,
    f: float = WGS84_FLATTENING,
) -> tuple:
    """
    Convert geodetic coordinates (LLH) to ECEF.

    :param float lat: lat in degrees
    :param float lon: lon on degrees
    :param float height: ellipsoidal height in metres
    :param float a: semi-major axis (6378137.0 for WGS84)
    :param float f: flattening (298.257223563 for WGS84)
    :return: tuple of ECEF (X, Y, Z) as floats
    :rtype: tuple
    """

    lat, lon = [c * pi / 180 for c in (lat, lon)]

    f = 1 / f
    e2 = f * (2 - f)
    a2 = a**2
    b2 = a2 * (1 - e2)

    N = a / sqrt(1 - e2 * sin(lat) ** 2)
    x = (N + height) * cos(lat) * cos(lon)
    y = (N + height) * cos(lat) * sin(lon)
    z = ((b2 / a2) * N + height) * sin(lat)

    return x, y, z


def llh2iso6709(lat: float, lon: float, alt: float, crs: str = WGS84) -> str:
    """
    Convert decimal degrees and alt to ISO6709 format
    e.g. "+27.5916+086.5640+8850CRSWGS_84/".

    :param float lat: latitude
    :param float lon: longitude
    :param float alt: altitude (hMSL)
    :param float crs: coordinate reference system (default = WGS_84)
    :return: position in ISO6709 format
    :rtype: str

    """

    lati, loni, alti = ["-" if c < 0 else "+" for c in (lat, lon, alt)]
    return f"{lati}{abs(lat)}{loni}{abs(lon)}{alti}{alt}CRS{crs}/"


def msgdesc(msgID: str) -> str:
    """
    Return descriptive string for NMEA msgId.

    :param msgID str: message ID e.g. 'GGA'
    :return: description of message
    :rtype: str

    """
    # pylint: disable=invalid-name

    if msgID in NMEA_MSGIDS:
        return NMEA_MSGIDS[msgID]
    if msgID in NMEA_MSGIDS_PROP:
        return NMEA_MSGIDS_PROP[msgID]
    return f"Unknown msgID {msgID}"


def planar(
    lat1: float,
    lon1: float,
    lat2: float,
    lon2: float,
    radius: int = WGS84_SMAJ_AXIS,
) -> float:
    """
    Calculate planar distance between two coordinates using planar
    approximation formula.

    NB suitable for coordinates less than around 50m apart. For
    larger separations, use the haversine great circle formula.

    :param float lat1: lat1
    :param float lon1: lon1
    :param float lat2: lat2
    :param float lon2: lon2
    :param float radius: radius in m (Earth = 6378137 m)
    :return: planar distance in m
    :rtype: float
    """

    phi1, lambda1, phi2, lambda2 = [c * pi / 180 for c in (lat1, lon1, lat2, lon2)]
    dlambda = (lambda2 - lambda1) * cos(phi1)
    dphi = phi2 - phi1
    dist = radius * sqrt(dlambda * dlambda + dphi * dphi)

    return dist


def time2str(tim: datetime.time) -> str:
    """
    Convert datetime.time to NMEA formatted string.

    :param datetime.time tim: time
    :return: NMEA formatted time string hhmmss.ss
    :rtype: str
    """

    try:
        return tim.strftime("%H%M%S.%f")[0:9]
    except (AttributeError, TypeError, ValueError):
        return ""


def time2utc(times: str) -> datetime.time:
    """
    Convert NMEA Time to UTC datetime.

    :param str times: NMEA time hhmmss.ss
    :return: UTC time hh:mm:ss.ss
    :rtype: datetime.time
    """

    try:
        if len(times) == 6:  # decimal seconds is omitted
            times = times + ".00"
        utc = datetime.strptime(times, "%H%M%S.%f")
        return utc.time()
    except (TypeError, ValueError):
        return ""



"""
NMEA Protocol Proprietary Set payload definitions

THESE ARE THE PAYLOAD DEFINITIONS FOR PROPRIETARY _SET_ MESSAGES _TO_
THE RECEIVER (e.g. Configuration commands).

NB: Attribute names must be unique within each message id.
NB: Avoid reserved names 'msgID', 'talker', 'payload', 'checksum'.

NB: Repeating groups must be defined as a tuple thus
    'group': ('numr', {dict})
    where
    - 'numr' is either:
       a) an integer representing a fixed number of repeats e.g 32
       b) a string representing the name of a preceding attribute
          containing the number of repeats e.g. 'numCh'
       c) 'None' for an indeterminate repeating group
          (only one such group is permitted per message type)
    - {dict} is the nested dictionary containing the repeating
      attributes

Created on 19 Aug 2024

While the NMEA 0183 © protocol is proprietary, the information here
has been collated from public domain sources.

:author: semuadmin (Steve Smith)
"""

from pynmeagps.nmeatypes_core import CH, DE, DT, HX, IN, LA, LAD, LN, LND, QS, ST, TM
from pynmeagps.nmeatypes_get_prop import NMEA_PAYLOADS_GET_PROP

NMEA_PAYLOADS_SET_PROP = {
    # ***************************************************************
    # Quectel LC29H / LC79H Proprietary SET message types
    # https://www.quectel.com/download/quectel_lc29hlc79h_series_gnss_protocol_specification_v1-5/
    # quectel_lc29hlc79h_series_gnss_protocol_specification_v1-5.pdf
    # ***************************************************************
    "AIR002": {},
    "AIR003": {},
    "AIR004": {},
    "AIR005": {},
    "AIR006": {},
    "AIR007": {},
    "AIR050": {
        "time": IN,
    },
    "AIR058": {
        "minsnr": IN,
    },
    "AIR062": {
        "type": IN,
        "rate": IN,
    },
    "AIR066": {
        "gpsEnabled": IN,
        "glonassEnabled": IN,
        "galileoEnabled": IN,
        "bdsEnabled": IN,
        "qzssEnabled": IN,
        "navicEnabled": IN,
    },
    "AIR070": {
        "speedThreshold": IN,
    },
    "AIR072": {
        "degree": IN,
    },
    "AIR074": {
        "enabled": IN,
    },
    "AIR080": {
        "navMode": IN,
    },
    "AIR086": {
        "status": IN,
    },
    "AIR100": {
        "nmeaMode": IN,
        "reserved": IN,
    },
    "AIR104": {
        "dualBandEnabled": IN,
    },
    "AIR382": {
        "enabled": IN,
    },
    "AIR391": {
        "cmdType": IN,
    },
    "AIR400": {
        "mode": IN,
    },
    "AIR410": {
        "enabled": IN,
    },
    "AIR420": {
        "enabled": IN,
    },
    "AIR432": {
        "mode": IN,
    },
    "AIR434": {
        "enabled": IN,
    },
    "AIR436": {
        "enabled": IN,
    },
    "AIR437": {},
    "AIR511": {},
    "AIR513": {},
    "AIR650": {
        "second": IN,
    },
    "AIR752": {
        "ppsType": IN,
    },
    "AIR864": {
        "portType": IN,  # 0 = UART
        "portIndex": IN,  # 0 = UART1
        "baudrate": IN,
    },
    "AIR866": {
        "portType": IN,  # 0 = UART
        "portIndex": IN,  # 0 = UART1
        "flowControl": IN,
    },
    "AIR6010": {
        "type": IN,
        "rate": IN,
    },
    # *********************************************
    # GARMIN PROPRIETARY MESSAGES
    # *********************************************
    "GRMI": {  # sensor initialisation information
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "date": DT,
        "time": TM,
        "rcvr_cmd": CH,
    },
    "GRMC": {  # sensor configuration information
        "fix": CH,
        "alt": DE,
        "dtm": ST,
        "smAxis": DE,
        "iffac": DE,
        "xecc": DE,
        "yecc": DE,
        "zecc": DE,
        "diff": CH,
        "baudRate": IN,
        "vfilt": IN,
        "reserved1": ST,
        "reserved2": ST,
        "drtime": IN,
    },
    "GRMC1": {  # additional sensor configuration information
        "nmeatim": IN,
        "bphase": IN,
        "autopos": IN,
        "dgpsfr": DE,
        "dgpsbr": IN,
        "dgpssc": IN,
        "nmeaver": IN,
        "dgpsmod": CH,
        "pwrsave": CH,
        "attran": IN,
        "autopwr": IN,
        "extpwr": IN,
    },
    "GRMO": {  # output sentence enable
        "msgId": ST,
        "tgtmode": IN,
    },
    "GRMW": {  # additional waypoint information
        "wptId": ST,
        "alt": DE,
        "symnum": HX,
        "comment": ST,
    },
    # ***************************************************************
    # Locosys Proprietary Messages SET
    # https://www.locosystech.com/Templates/att/LS2303x-UDG_datasheet_v0.6.pdf?lng=en
    # ***************************************************************
    # Proprietary Locosys command sets not yet implemented, e.g.
    # Perform a Cold start $PMTK103*30
    # Perform a Warm start $PMTK102*31
    # Perform a Hot start $PMTK101*32
    # Perform a Full Cold start $PMTK104*37
    # Disable GLL message $PMTK314,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0*29
    # Disable GSV message $PMTK314,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29
    # Disable GLL & GSV message $PMTK314,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0*28
    # Factory default output message $PMTK314,-1*04
    # Navigate with GPS+GALILEO $PMTK353,1,0,1,0,0*2B
    # Navigate with GPS+GLONASS+GALILEO $PMTK353,1,1,1,0,0*2A
    # Navigate with GPS+BEIDOU $PMTK353,1,0,0,0,1*2B
    # Entering Standby Mode1 $PMTK161,0*28
    "INVCRES": {
        "value": IN,  # 0 = clear data
    },
    "INVCSTR": {
        "value": IN,  # 14 = start session
    },
    "INVMSLOPE": {
        "status": IN,  # 0 = disable, 1 =  enable
    },
    "LSC": {
        "msgType": ST,  # e.g. "SLOPE", "MEMS", "ATTIT", "VER"
        "value": IN,  # e.g. 0 (not present if msgType == "VER")
    },
    # ***************************************************************
    # Quectel LG290P Proprietary message types
    # https://quectel.com/content/uploads/2024/09/Quectel_LG290P03_GNSS_Protocol_Specification_V1.0.pdf
    # ***************************************************************
    "QTMBKP": {
        "second": IN,
    },
    "QTMCFGAIC": NMEA_PAYLOADS_GET_PROP["QTMCFGAIC"],
    "QTMCFGDR": NMEA_PAYLOADS_GET_PROP["QTMCFGDR"],
    "QTMCFGCNST": NMEA_PAYLOADS_GET_PROP["QTMCFGCNST"],
    "QTMCFGDRRTD": NMEA_PAYLOADS_GET_PROP["QTMCFGDRRTD"],
    "QTMCFGDRHOT": NMEA_PAYLOADS_GET_PROP["QTMCFGDRHOT"],
    "QTMCFGEINSMSG": NMEA_PAYLOADS_GET_PROP["QTMCFGEINSMSG"],
    "QTMCFGFIXRATE": NMEA_PAYLOADS_GET_PROP["QTMCFGFIXRATE"],
    "QTMCFGGEOFENCE": NMEA_PAYLOADS_GET_PROP["QTMCFGGEOFENCE"],
    "QTMCFGGEOFENCE_POLY": NMEA_PAYLOADS_GET_PROP["QTMCFGGEOFENCE_POLY"],
    "QTMCFGGEOFENCE_DIS": NMEA_PAYLOADS_GET_PROP["QTMCFGGEOFENCE_DIS"],
    "QTMCFGIMUINT": NMEA_PAYLOADS_GET_PROP["QTMCFGIMUINT"],
    "QTMCFGIMUTC": NMEA_PAYLOADS_GET_PROP["QTMCFGIMUTC"],
    "QTMCFGLA": NMEA_PAYLOADS_GET_PROP["QTMCFGLA"],
    "QTMCFGLAM": NMEA_PAYLOADS_GET_PROP["QTMCFGLAM"],
    "QTMCFGMSGRATE_NOVER": NMEA_PAYLOADS_GET_PROP["QTMCFGMSGRATE_NOVER"],
    "QTMCFGMSGRATE": NMEA_PAYLOADS_GET_PROP["QTMCFGMSGRATE"],
    "QTMCFGNAVMODE": NMEA_PAYLOADS_GET_PROP["QTMCFGNAVMODE"],
    "QTMCFGNMEADP": NMEA_PAYLOADS_GET_PROP["QTMCFGNMEADP"],
    "QTMCFGNMEATID": NMEA_PAYLOADS_GET_PROP["QTMCFGNMEATID"],
    "QTMCFGODO": NMEA_PAYLOADS_GET_PROP["QTMCFGODO"],
    "QTMCFGPPS": NMEA_PAYLOADS_GET_PROP["QTMCFGPPS"],
    "QTMCFGPPS_DIS": {
        "status": QS,
        "ppsindex": IN,
        "enable": IN,  # should be 0
    },
    "QTMCFGPROT": NMEA_PAYLOADS_GET_PROP["QTMCFGPROT"],
    "QTMCFGRCVRMODE": NMEA_PAYLOADS_GET_PROP["QTMCFGRCVRMODE"],
    "QTMCFGRSID": NMEA_PAYLOADS_GET_PROP["QTMCFGRSID"],
    "QTMCFGRTCM": NMEA_PAYLOADS_GET_PROP["QTMCFGRTCM"],
    "QTMCFGRTK": NMEA_PAYLOADS_GET_PROP["QTMCFGRTK"],
    "QTMCFGRTKSRCTYPE": NMEA_PAYLOADS_GET_PROP["QTMCFGRTKSRCTYPE"],
    "QTMCFGSAT": NMEA_PAYLOADS_GET_PROP["QTMCFGSAT"],
    "QTMCFGSAT_LOW": NMEA_PAYLOADS_GET_PROP["QTMCFGSAT_LOW"],
    "QTMCFGSIGNAL": NMEA_PAYLOADS_GET_PROP["QTMCFGSIGNAL"],
    "QTMCFGSTATICHOLD": NMEA_PAYLOADS_GET_PROP["QTMCFGSTATICHOLD"],
    "QTMCFGSVIN": NMEA_PAYLOADS_GET_PROP["QTMCFGSVIN"],
    # sets all parms for specified interface:
    "QTMCFGUART": NMEA_PAYLOADS_GET_PROP["QTMCFGUART"],
    "QTMCFGVEHMOT": NMEA_PAYLOADS_GET_PROP["QTMCFGVEHMOT"],
    # sets baud rate for specified interface
    "QTMCFGUART_BAUD": {
        "status": QS,
        "portid": IN,
        "baudrate": IN,
    },
    # sets all parms for current interface:
    "QTMCFGUART_CURR": {
        "status": QS,
        "baudrate": IN,
        "databit": IN,
        "parity": IN,
        "stopbit": IN,
        "flowctrl": IN,
    },
    # sets baud rate for current interface:
    "QTMCFGUART_CURRBAUD": {
        "status": QS,
        "baudrate": IN,
    },
    "QTMCOLD": {},
    "QTMDEBUGOFF": {},
    "QTMDEBUGON": {},
    "QTMDRCLR": {},
    "QTMDRSAVE": {},
    "QTMGNSSSTART": {},
    "QTMGNSSSTOP": {},
    "QTMHOT": {},
    "QTMRESETODO": {},
    "QTMRESTOREPAR": {},
    "QTMSAVEPAR": {},
    "QTMSRR": {},
    "QTMUNIQID": {},
    "QTMVERNO": {},
    "QTMWARM": {},
    # *********************************************
    # U-BLOX PROPRIETARY MESSAGES
    # *********************************************
    "UBX40": {  # set message rates per port
        "msgId": ST,  # '40'
        "id": ST,  # e.g. GLL
        "rddc": IN,  # I2C
        "rus1": IN,  # UART1
        "rus2": IN,  # UART2
        "rusb": IN,  # USB
        "rspi": IN,  # SPI
        "reserved": IN,
    },
    "UBX41": {  # configure port protocols
        "msgId": ST,  # '41'
        "portId": IN,
        "inProto": HX,
        "outProto": HX,
        "baudRate": IN,
        "autobauding": IN,
    },
}



"""
NMEA Custom Exception Types

Created on 04 Mar 2021

:author: semuadmin (Steve Smith)
:copyright: semuadmin © 2020
:license: BSD 3-Clause
"""


class NMEAParseError(Exception):
    """
    NMEA Parsing error.
    """


class NMEAStreamError(Exception):
    """
    NMEA Streaming error.
    """


class NMEAMessageError(Exception):
    """
    NMEA Undefined message class/id.
    Essentially a prompt to add missing payload types to UBX_PAYLOADS.
    """


class NMEATypeError(Exception):
    """
    NMEA Undefined payload attribute type.
    Essentially a prompt to fix incorrect payload definitions to UBX_PAYLOADS.
    """



"""
Quectel LG290P Stream method tests for pynmeagps

Created on 19 Aug 2024

*** NB: must be saved in UTF-8 format ***

:author: semuadmin (Steve Smith)
"""

import os
import sys
import unittest

from pynmeagps import (
    NMEAReader,
    NMEAMessage,
    VALCKSUM,
    VALMSGID,
    SET,
    GET,
    POLL,
    hex2str,
)

DIRNAME = os.path.dirname(__file__)


class QuectelStreamTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def tearDown(self):
        pass

    def catchio(self):
        """
        Capture stdout as string.
        """

        self._saved_stdout = sys.stdout
        self._strout = os.StringIO()
        sys.stdout = self._strout

    def restoreio(self) -> str:
        """
        Return captured output and restore stdout.
        """

        sys.stdout = self._saved_stdout
        return self._strout.getvalue().strip()

    def testNMEAQUECTELGET(self):  # test proprietary Quectel LG290P GET messages
        EXPECTED_RESULTS = (
            "<NMEA(PQTMVER, msgver=1, vername=MODULE, verstr=LG290P03AANR01A03S, builddate=2024/04/30, buildtime=10:53:07)>",
            "<NMEA(PQTMVERNO, verstr=LG290P03AANR01A03S, builddate=2024/04/30, buildtime=10:53:07)>",
            "<NMEA(PQTMEPE, msgver=2, epenorth=1.0, epeeast=1.0, epedown=1.0, epe2d=1.414, epe3d=1.732)>",
            "<NMEA(PQTMVEL, msgver=1, time=15:45:12.100000, veln=1.251, vele=2.452, veld=1.245, gndspd=2.752, spd=3.021, hdg=180.512, gndspdacc=0.124, spdacc=0.254, hdgacc=0.25)>",
            "<NMEA(PQTMGEOFENCESTATUS, msgver=1, time=12:45:21, staten_01=1, staten_02=2, staten_03=2, staten_04=2)>",
            "<NMEA(PQTMTXT, msgver=1, totalsennum=1, sennum=1, textid=1, text=0x105f0cf810417c00)>",
            "<NMEA(PQTMSVINSTATUS, msgver=1, tow=1000, valid=1, reserved0=, reserved1=1, obs=20, cfgdur=100, meanx=-2484434.3645, meany=4875976.9741, meanz=3266161.3412, meanacc=1.2415)>",
            "<NMEA(PQTMPVT, msgver=1, tow=1000, date=20221225, time=16:33:55, reserved=, fixtype=0, numsv=0, leaps=, lat=, lon=, alt=, sep=, veln=, vele=, veld=, spd=, hdg=, hdop=99.99, pdop=99.99)>",
            "<NMEA(PQTMPVT, msgver=1, tow=31075000, date=20221225, time=08:37:37, reserved=, fixtype=3, numsv=9, leaps=18, lat=31.12738291, lon=117.2637291, alt=34.212, sep=5.267, veln=3.212, vele=2.928, veld=0.238, spd=4.346, hdg=34.12, hdop=2.16, pdop=4.38)>",
            "<NMEA(PQTMDOP, msgver=1, tow=570643000, gdop=1.01, pdop=0.88, tdop=0.49, vdop=0.73, hdop=0.5, ndop=0.36, edop=0.35)>",
            "<NMEA(PQTMDOP, msgver=1, tow=, gdop=99.99, pdop=99.99, tdop=99.99, vdop=99.99, hdop=99.99, ndop=99.99, edop=99.99)>",
            "<NMEA(PQTMPL, msgver=1, tow=55045200, pul=5.0, reserved1=1, reserved2=1, plposn=2879.0, plpose=2718.0, plposd=4766.0, plveln=5344.0, plvele=4323.0, plveld=10902.0, reserved3=, reserved4=, pltime=)>",
            "<NMEA(PQTMODO, msgver=1, time=12:06:35, state=1, dist=112.3)>",
            "<NMEA(PQTMLS, msgver=1, tow=195494, lsRef=1, wno=2299, leapsecond=18, flag=0, lsfRef=1, reserved=, wnlsf=137, dn=7, lsf=18)>",
            "<NMEA(PQTMQVER, status=OK, msgver=1, description=MODULE, version=LC29HCANR11A07S_DSA4, buildDate=2025/06/03, buildTime=15:28:01)>",
            "<NMEA(PQTMJAMMINGSTATUS, msgver=1, jammingstatus=1)>",
            "<NMEA(PQTMSTD, reserved=1, utc=13:11:37, wno=2368, tow=393115, stdLat=5.36, stdLon=4.62, stdAlt=12.12, stdSep=, stdVelN=0.13, stdVelE=0.13, stdVelD=0.14, stdSpd=0.23, stdRoll=, stdPitch=, stdHeading=)>",
            "<NMEA(PQTMSTD, reserved=1, utc=13:15:23.314000, wno=0, tow=393341, stdLat=, stdLon=, stdAlt=, stdSep=, stdVelN=, stdVelE=, stdVelD=, stdSpd=, stdRoll=, stdPitch=, stdHeading=)>",
            "<NMEA(PQTMGETUTC, status=OK, year=2024, month=10, day=22, hour=2, minute=52, second=30, millisecond=295, reserved=, leapsecond=18)>",
            "<NMEA(PQTMSN, status=OK, snid=1, length=16, serialno=1234567890ABCDEF)>",
            "<NMEA(PQTMDRCAL, msgver=1, calstate=0, navtype=1)>",
            "<NMEA(PQTMIMUTYPE, msgver=1, imustatus=2)>",
            "<NMEA(PQTMVEHMSG, msgver=2, timestamp=153954, vehicle_info_01=100.0, vehicle_info_02=1.0)>",
            "<NMEA(PQTMINS, timestamp=240951, soltype=1, lat=31.82222216, lon=117.11578436, height=62.555605, velN=-0.004233, velE=0.005535, velD=-0.004011, roll=0.0, pitch=0.0, yaw=127.41)>",
            "<NMEA(PQTMIMU, timestamp=45454, accX=-1.35673, accY=-0.210568, accZ=9.75793, angrateX=0.564879, angrateY=0.549612, angrateZ=-0.412209, wheeltick=0, last_tick_timestamp=0)>",
            "<NMEA(PQTMGPS, timestamp=86139, tow=94183, lat=31.82218794, lon=117.11579022, alt=65.75508, speed=0.027, heading=94.68, accuracy=2.533952, HDOP=0.555471, PDOP=0.886183, numSv=29, fixMode=3)>",
            "<NMEA(PQTMCFGEINSMSG, type=1, ins_enabled=1, imu_enabled=1, gps_enabled=1, rate=10)>",
            "<NMEA(PQTMVEHMOT, msgver=1, vehicle_info_01=0.288124, vehicle_info_02=0.15993)>",
            "<NMEA(PQTMSENMSG, msgver=2, timestamp=1000, sensor_info_01=22.21, sensor_info_02=0.124521, sensor_info_03=1.241541, sensor_info_04=0.912451, sensor_info_05=0.145785, sensor_info_06=1.241541, sensor_info_07=8.954214)>",
            "<NMEA(PQTMSENMSG, msgver=4, timestamp=1000, sensor_info_01=22.21, sensor_info_02=0.124521, sensor_info_03=1.241541, sensor_info_04=0.912451, sensor_info_05=0.145785, sensor_info_06=1.241541, sensor_info_07=8.954214)>",
            "<NMEA(PQTMCFGDRRTD, status=OK, time=600, distance=10000)>",
            "<NMEA(PQTMCFGIMUTC, status=OK, imustatus=1)>",
            "<NMEA(PQTMDRPVA, msgver=1, timestamp=1000, time=163355.0, solType=0, lat=, lon=, alt=, sep=, velN=, velE=, velD=, speed=, roll=, pitch=, heading=)>",
            "<NMEA(PQTMDRPVA, msgver=1, timestamp=75000, time=83737.0, solType=2, lat=31.12738291, lon=117.2637291, alt=34.212, sep=5.267, velN=3.212, velE=2.928, velD=0.238, speed=4.346, roll=0.392663, pitch=1.300793, heading=0.030088)>",
            "<NMEA(PQTMCFGDRHOT, status=OK, mode=1)>",
            "<NMEA(PQTMCFGDR, status=OK, drstate=1)>",
            "<NMEA(PQTMCFGIMUINT, status=OK, index=1, mode=1, reserved=0, actLevel=1)>",
            "<NMEA(PQTMCFGLA, status=OK, type=1, laX=0.212, laY=0.514, laZ=0.113)>",
            "<NMEA(PQTMCFGLAM, status=OK, mode=2)>",
            "<NMEA(PQTMCFGRTKSRCTYPE, status=OK, srctype=1)>",
            "<NMEA(PQTMCFGSTATICHOLD, status=OK, shstate=1)>",
            "<NMEA(PQTMCFGVEHMOT, status=OK, mode=1, vehtype=1)>",
            "<NMEA(PQTMDRCLR, status=OK)>",
            "<NMEA(PQTMDRSAVE, status=OK)>",
            "<NMEA(PQTMVEHATT, msgver=1, timestamp=1000, roll=10.002154, pitch=20.235412, heading=160.145185, acc_Roll=1.254123, acc_Pitch=5.451214, acc_Heading=5.102154)>",
        )
        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "quectel_nmea_get.log"), "rb") as stream:
            nmr = NMEAReader(
                stream, validate=VALCKSUM | VALMSGID, quitonerror=2, msgmode=GET
            )
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, len(EXPECTED_RESULTS))

    def testNMEAQUECTELPOLL(self):  # test proprietary Quectel LG290P POLL messages
        EXPECTED_RESULTS = (
            "<NMEA(PQTMCFGUART, status=R)>",
            "<NMEA(PQTMCFGUART, status=R, portid=1)>",
            "<NMEA(PQTMCFGPPS, status=R, ppsindex=1)>",
            "<NMEA(PQTMCFGPROT, status=R, porttype=1, portid=1)>",
            "<NMEA(PQTMCFGNMEADP, status=R)>",
            "<NMEA(PQTMCFGMSGRATE, status=R, msgname=GGA)>",
            "<NMEA(PQTMCFGMSGRATE, status=R, msgname=PQTMEPE, msgver=2)>",
            "<NMEA(PQTMCFGMSGRATE, status=R, msgname=RTCM3-1005)>",
            "<NMEA(PQTMCFGMSGRATE, status=R, msgname=RTCM3-107X)>",
            "<NMEA(PQTMCFGMSGRATE, status=R, msgname=RTCM3-1019)>",
            "<NMEA(PQTMCFGGEOFENCE, status=R, geofenceindex=0)>",
            "<NMEA(PQTMCFGSVIN, status=R)>",
            "<NMEA(PQTMCFGRCVRMODE, status=R)>",
            "<NMEA(PQTMCFGFIXRATE, status=R)>",
            "<NMEA(PQTMCFGRTK, status=R)>",
            "<NMEA(PQTMCFGCNST, status=R)>",
            "<NMEA(PQTMCFGODO, status=R)>",
            "<NMEA(PQTMCFGSIGNAL, status=R)>",
            "<NMEA(PQTMCFGSAT, status=R, systemid=1, signalid=1)>",
            "<NMEA(PQTMCFGSAT, status=R, systemid=4, signalid=1)>",
            "<NMEA(PQTMCFGRSID, status=R)>",
            "<NMEA(PQTMGETUTC)>",
            "<NMEA(PQTMSN)>",
            "<NMEA(PQTMCFGAIC, status=R)>",
            "<NMEA(PQTMCFGNAVMODE, status=R)>",
            "<NMEA(PQTMCFGNMEATID, status=R)>",
            "<NMEA(PQTMQVER, msgver=1)>",
            "<NMEA(PQTMCFGDR, status=R)>",
            "<NMEA(PQTMCFGIMUINT, status=R)>",
            "<NMEA(PQTMCFGLA, status=R)>",
            "<NMEA(PQTMCFGLAM, status=R)>",
            "<NMEA(PQTMCFGRTKSRCTYPE, status=R)>",
            "<NMEA(PQTMCFGSTATICHOLD, status=R)>",
            "<NMEA(PQTMCFGVEHMOT, status=R)>",
        )
        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "quectel_nmea_poll.log"), "rb") as stream:
            nmr = NMEAReader(
                stream, validate=VALCKSUM | VALMSGID, quitonerror=2, msgmode=POLL
            )
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, len(EXPECTED_RESULTS))

    def testNMEAQUECTELCOMMAND(
        self,
    ):  # test proprietary Quectel LG290P Command (SET) messages
        EXPECTED_RESULTS = (
            "<NMEA(PQTMCOLD)>",
            "<NMEA(PQTMWARM)>",
            "<NMEA(PQTMHOT)>",
            "<NMEA(PQTMSRR)>",
            "<NMEA(PQTMUNIQID)>",
            "<NMEA(PQTMSAVEPAR)>",
            "<NMEA(PQTMRESTOREPAR)>",
            "<NMEA(PQTMVERNO)>",
            "<NMEA(PQTMCFGUART, status=W, baudrate=115200)>",
            "<NMEA(PQTMCFGUART, status=W, portid=1, baudrate=115200)>",
            "<NMEA(PQTMCFGUART, status=W, baudrate=115200, databit=8, parity=0, stopbit=1, flowctrl=0)>",
            "<NMEA(PQTMCFGUART, status=W, portid=1, baudrate=115200, databit=8, parity=0, stopbit=1, flowctrl=0)>",
            "<NMEA(PQTMCFGPPS, status=W, ppsindex=1, enable=1, duration=100, ppsmode=1, polarity=1, reserved=0)>",
            "<NMEA(PQTMCFGPPS, status=W, ppsindex=1, enable=0)>",
            "<NMEA(PQTMCFGPROT, status=W, porttype=1, portid=1, inputprot=00000005, outputprot=00000005)>",
            "<NMEA(PQTMCFGNMEADP, status=W, utcdp=3, posdp=8, altdp=3, dopdp=2, spddp=3, cogdp=2)>",
            "<NMEA(PQTMCFGMSGRATE, status=W, msgname=GGA, rate=1)>",
            "<NMEA(PQTMCFGMSGRATE, status=W, msgname=PQTMEPE, rate=1, msgver=2)>",
            "<NMEA(PQTMCFGMSGRATE, status=W, msgname=RTCM3-1005, rate=1)>",
            "<NMEA(PQTMCFGMSGRATE, status=W, msgname=RTCM3-107X, rate=1, msgver=0)>",
            "<NMEA(PQTMCFGMSGRATE, status=W, msgname=RTCM3-1019, rate=1)>",
            "<NMEA(PQTMCFGGEOFENCE, status=W, geofenceindex=0, geofencemode=0)>",
            "<NMEA(PQTMGNSSSTART)>",
            "<NMEA(PQTMGNSSSTOP)>",
            "<NMEA(PQTMCFGSVIN, status=W, svinmode=1, cfgcnt=3600, acclimit=1.2, ecefx=-2519265.0514, ecefy=4849534.9045, ecefz=3277834.6432)>",
            "<NMEA(PQTMCFGRCVRMODE, status=W, rcvrmode=2)>",
            "<NMEA(PQTMDEBUGON)>",
            "<NMEA(PQTMDEBUGOFF)>",
            "<NMEA(PQTMCFGFIXRATE, status=W, fixinterval=1000)>",
            "<NMEA(PQTMCFGRTK, status=W, diffmode=1, relmode=1)>",
            "<NMEA(PQTMCFGCNST, status=W, gps=1, glonass=1, galileo=1, beidou=1, qzss=0, navic=0)>",
            "<NMEA(PQTMCFGODO, status=W, state=1, initdist=10.5)>",
            "<NMEA(PQTMRESETODO)>",
            "<NMEA(PQTMCFGSIGNAL, status=W, gpssig=7, glonasssig=3, galileosig=F, beidousig=3F, qzsssig=7, navicsig=1)>",
            "<NMEA(PQTMCFGSAT, status=W, systemid=1, signalid=1, masklow=FFFFFFFF)>",
            "<NMEA(PQTMCFGSAT, status=W, systemid=4, signalid=1, masklow=BFFCBFFF, maskhigh=1C003FFF)>",
            "<NMEA(PQTMCFGRSID, status=W, rsid=1024)>",
            "<NMEA(PQTMCFGRTCM, status=W, msmtype=4, msmmode=0, msmelevthd=-90, reserved1=07, reserved2=06, ephmode=1, ephinterval=0)>",
            "<NMEA(PQTMBKP, second=66)>",
            "<NMEA(PQTMCFGAIC, status=W, state=1)>",
            "<NMEA(PQTMCFGNAVMODE, status=W, mode=10)>",
            "<NMEA(PQTMCFGNMEATID, status=W, mainTalker=GP, gsvTalker=0)>",
            "<NMEA(PQTMCFGDR, status=W, drstate=1)>",
            "<NMEA(PQTMCFGIMUINT, status=W, index=1, mode=1, reserved=0, actLevel=1)>",
            "<NMEA(PQTMCFGLA, status=W, type=1, laX=0.212, laY=0.514, laZ=0.113)>",
            "<NMEA(PQTMCFGLAM, status=W, mode=2)>",
            "<NMEA(PQTMCFGRTKSRCTYPE, status=W, srctype=1)>",
            "<NMEA(PQTMCFGSTATICHOLD, status=W, shstate=1)>",
            "<NMEA(PQTMCFGVEHMOT, status=W, mode=1, vehtype=1)>",
            "<NMEA(PQTMDRCLR)>",
            "<NMEA(PQTMDRSAVE)>",
        )
        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "quectel_nmea_command.log"), "rb") as stream:
            nmr = NMEAReader(
                stream, validate=VALCKSUM | VALMSGID, quitonerror=2, msgmode=SET
            )
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, len(EXPECTED_RESULTS))

    def testNMEAQUECTELRESPONSE(
        self,
    ):  # test proprietary Quectel LG290P Response (GET) messages
        EXPECTED_RESULTS = (
            "<NMEA(PQTMVER, msgver=1, vername=MODULE, verstr=LG290P03AANR01A03S, builddate=2024/04/30, buildtime=10:53:07)>",
            "<NMEA(PQTMUNIQID, status=OK, length=16, ID=81D62010EE0AF375BDF5952CDC3757A1)>",
            "<NMEA(PQTMSAVEPAR, status=OK)>",
            "<NMEA(PQTMRESTOREPAR, status=OK)>",
            "<NMEA(PQTMVERNO, verstr=LG290P03AANR01A03S, builddate=2024/04/30, buildtime=10:53:07)>",
            "<NMEA(PQTMCFGUART, status=OK)>",
            "<NMEA(PQTMCFGUART, status=OK)>",
            "<NMEA(PQTMCFGUART, status=OK)>",
            "<NMEA(PQTMCFGUART, status=OK)>",
            "<NMEA(PQTMCFGUART, status=OK, portid=1, baudrate=115200, databit=8, parity=0, stopbit=1, flowctrl=0)>",
            "<NMEA(PQTMCFGUART, status=OK, portid=1, baudrate=115200, databit=8, parity=0, stopbit=1, flowctrl=0)>",
            "<NMEA(PQTMCFGPPS, status=OK)>",
            "<NMEA(PQTMCFGPPS, status=OK, ppsindex=1, enable=1, duration=100, ppsmode=1, polarity=1, reserved=0)>",
            "<NMEA(PQTMCFGPPS, status=OK)>",
            "<NMEA(PQTMCFGPROT, status=OK)>",
            "<NMEA(PQTMCFGPROT, status=OK, porttype=1, portid=1, inputprot=00000005, outputprot=00000005)>",
            "<NMEA(PQTMCFGNMEADP, status=OK)>",
            "<NMEA(PQTMCFGNMEADP, status=OK, utcdp=3, posdp=8, altdp=3, dopdp=2, spddp=3, cogdp=2)>",
            "<NMEA(PQTMEPE, msgver=2, epenorth=1.0, epeeast=1.0, epedown=1.0, epe2d=1.414, epe3d=1.732)>",
            "<NMEA(PQTMCFGMSGRATE, status=OK)>",
            "<NMEA(PQTMCFGMSGRATE, status=OK, msgname=GGA, rate=1, msgver=)>",
            "<NMEA(PQTMCFGMSGRATE, status=OK)>",
            "<NMEA(PQTMCFGMSGRATE, status=OK, msgname=PQTMEPE, rate=1, msgver=2)>",
            "<NMEA(PQTMCFGMSGRATE, status=OK)>",
            "<NMEA(PQTMCFGMSGRATE, status=OK, msgname=RTCM3-1005, rate=1)>",
            "<NMEA(PQTMCFGMSGRATE, status=OK)>",
            "<NMEA(PQTMCFGMSGRATE, status=OK, msgname=RTCM3-107X, rate=1, msgver=0)>",
            "<NMEA(PQTMCFGMSGRATE, status=OK)>",
            "<NMEA(PQTMCFGMSGRATE, status=OK, msgname=RTCM3-1019, rate=1)>",
            "<NMEA(PQTMVEL, msgver=1, time=15:45:12.100000, veln=1.251, vele=2.452, veld=1.245, gndspd=2.752, spd=3.021, hdg=180.512, gndspdacc=0.124, spdacc=0.254, hdgacc=0.25)>",
            "<NMEA(PQTMCFGGEOFENCE, status=OK)>",
            "<NMEA(PQTMCFGGEOFENCE, status=OK)>",
            "<NMEA(PQTMCFGGEOFENCE, status=OK, geofenceindex=0, geofencemode=1, reserved=0, shape=0, lat0=31.451248, lon0=117.451245, radiuslat1=100.5)>",
            "<NMEA(PQTMCFGGEOFENCE, status=OK, geofenceindex=0, geofencemode=0)>",
            "<NMEA(PQTMGEOFENCESTATUS, msgver=1, time=12:45:21, staten_01=1, staten_02=2, staten_03=2, staten_04=2)>",
            "<NMEA(PQTMGNSSSTART, status=OK)>",
            "<NMEA(PQTMGNSSSTOP, status=OK)>",
            "<NMEA(PQTMTXT, msgver=1, totalsennum=1, sennum=1, textid=1, text=0x105f0cf810417c00)>",
            "<NMEA(PQTMCFGSVIN, status=OK)>",
            "<NMEA(PQTMCFGSVIN, status=OK, svinmode=1, cfgcnt=3600, acclimit=1.2, ecefx=-2519265.0514, ecefy=4849534.9045, ecefz=3277834.6432)>",
            "<NMEA(PQTMSVINSTATUS, msgver=1, tow=1000, valid=1, reserved0=, reserved1=1, obs=20, cfgdur=100, meanx=-2484434.3645, meany=4875976.9741, meanz=3266161.3412, meanacc=1.2415)>",
            "<NMEA(PQTMPVT, msgver=1, tow=1000, date=20221225, time=16:33:55, reserved=, fixtype=0, numsv=0, leaps=, lat=, lon=, alt=, sep=, veln=, vele=, veld=, spd=, hdg=, hdop=99.99, pdop=99.99)>",
            "<NMEA(PQTMPVT, msgver=1, tow=31075000, date=20221225, time=08:37:37, reserved=, fixtype=3, numsv=9, leaps=18, lat=31.12738291, lon=117.2637291, alt=34.212, sep=5.267, veln=3.212, vele=2.928, veld=0.238, spd=4.346, hdg=34.12, hdop=2.16, pdop=4.38)>",
            "<NMEA(PQTMCFGRCVRMODE, status=OK)>",
            "<NMEA(PQTMCFGRCVRMODE, status=OK, rcvrmode=2)>",
            "<NMEA(PQTMDEBUGON, status=OK)>",
            "<NMEA(PQTMDEBUGOFF, status=OK)>",
            "<NMEA(PQTMCFGFIXRATE, status=OK)>",
            "<NMEA(PQTMCFGFIXRATE, status=OK, fixinterval=1000)>",
            "<NMEA(PQTMCFGRTK, status=OK)>",
            "<NMEA(PQTMCFGRTK, status=OK, diffmode=1, relmode=1)>",
            "<NMEA(PQTMCFGCNST, status=OK)>",
            "<NMEA(PQTMCFGCNST, status=OK, gps=1, glonass=1, galileo=1, beidou=1, qzss=0, navic=0)>",
            "<NMEA(PQTMDOP, msgver=1, tow=570643000, gdop=1.01, pdop=0.88, tdop=0.49, vdop=0.73, hdop=0.5, ndop=0.36, edop=0.35)>",
            "<NMEA(PQTMDOP, msgver=1, tow=, gdop=99.99, pdop=99.99, tdop=99.99, vdop=99.99, hdop=99.99, ndop=99.99, edop=99.99)>",
            "<NMEA(PQTMPL, msgver=1, tow=55045200, pul=5.0, reserved1=1, reserved2=1, plposn=2879.0, plpose=2718.0, plposd=4766.0, plveln=5344.0, plvele=4323.0, plveld=10902.0, reserved3=, reserved4=, pltime=)>",
            "<NMEA(PQTMCFGODO, status=OK)>",
            "<NMEA(PQTMCFGODO, status=ERROR, errcode=1)>",
            "<NMEA(PQTMCFGODO, status=OK, state=1, initdist=10.5)>",
            "<NMEA(PQTMRESETODO, status=OK)>",
            "<NMEA(PQTMODO, msgver=1, time=12:06:35, state=1, dist=112.3)>",
            "<NMEA(PQTMCFGSIGNAL, status=OK)>",
            "<NMEA(PQTMCFGSIGNAL, status=OK, gpssig=07, glonasssig=03, galileosig=0F, beidousig=3F, qzsssig=07, navicsig=01)>",
            "<NMEA(PQTMCFGSAT, status=OK)>",
            "<NMEA(PQTMCFGSAT, status=OK, systemid=1, signalid=01, masklow=FFFFFFFF)>",
            "<NMEA(PQTMCFGSAT, status=OK)>",
            "<NMEA(PQTMCFGSAT, status=ERROR, errcode=2)>",
            "<NMEA(PQTMCFGRSID, status=OK)>",
            "<NMEA(PQTMCFGRSID, status=OK, rsid=1024)>",
            "<NMEA(PQTMCFGRTCM, status=OK)>",
            "<NMEA(PQTMCFGGEOFENCE, status=R, geofenceindex=0, geofencemode=1, reserved=0, shape=3, lat0=31.451248, lon0=117.451245, radiuslat1=100.5, lon1=12.36, lat2=85.24, lon2=118.72, lat3=56.45, lon3=140.13)>",
            "<NMEA(PQTMCFGDR, status=OK)>",
            "<NMEA(PQTMCFGDR, status=OK, drstate=1)>",
            "<NMEA(PQTMCFGIMUINT, status=OK)>",
            "<NMEA(PQTMCFGIMUINT, status=OK, index=1, mode=1, reserved=0, actLevel=1)>",
            "<NMEA(PQTMCFGLA, status=OK)>",
            "<NMEA(PQTMCFGLA, status=OK, type=1, laX=0.212, laY=0.514, laZ=0.113)>",
            "<NMEA(PQTMCFGLAM, status=OK)>",
            "<NMEA(PQTMCFGLAM, status=OK, mode=2)>",
            "<NMEA(PQTMCFGRTKSRCTYPE, status=OK)>",
            "<NMEA(PQTMCFGRTKSRCTYPE, status=OK, srctype=1)>",
            "<NMEA(PQTMCFGSTATICHOLD, status=OK)>",
            "<NMEA(PQTMCFGSTATICHOLD, status=OK, shstate=1)>",
            "<NMEA(PQTMCFGVEHMOT, status=OK)>",
            "<NMEA(PQTMCFGVEHMOT, status=OK, mode=1, vehtype=1)>",
            "<NMEA(PQTMDRCLR, status=OK)>",
            "<NMEA(PQTMDRSAVE, status=OK)>",
        )
        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "quectel_nmea_response.log"), "rb") as stream:
            nmr = NMEAReader(
                stream, validate=VALCKSUM | VALMSGID, quitonerror=1, msgmode=GET
            )
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, len(EXPECTED_RESULTS))

    def testConstructors_QUECTEL(self):  # test Quectel constructors
        EXPECTED_PARSED = [
            "<NMEA(PQTMCFGPPS, status=R, ppsindex=0)>",
            "<NMEA(PQTMCFGPPS, status=OK, ppsindex=0, enable=0, duration=0, ppsmode=0, polarity=0, reserved=0)>",
            "<NMEA(PQTMCFGUART, status=W, baudrate=115200)>",
            "<NMEA(PQTMCFGUART, status=W, portid=1, baudrate=115200)>",
            "<NMEA(PQTMCFGUART, status=W, baudrate=115200, databit=8, parity=0, stopbit=1, flowctrl=0)>",
            "<NMEA(PQTMCFGUART, status=W, portid=1, baudrate=115200, databit=8, parity=0, stopbit=0, flowctrl=0)>",
            "<NMEA(PQTMCFGUART, status=R)>",
            "<NMEA(PQTMCFGUART, status=R, portid=1)>",
            "<NMEA(PQTMCFGMSGRATE, status=W, msgname=GLL, rate=1)>",
            "<NMEA(PQTMCFGMSGRATE, status=W, msgname=GLL, rate=1, msgver=2)>",
            "<NMEA(PQTMCFGMSGRATE, status=R, msgname=GLL)>",
            "<NMEA(PQTMCFGMSGRATE, status=R, msgname=PQTMEPE, msgver=2)>",
            "<NMEA(PQTMVERNO)>",
            "<NMEA(PQTMVER, msgver=1, vername=MODULE, verstr=LG290P03AANR01A03S, builddate=2024/04/30, buildtime=10:53:07)>",
            "<NMEA(PQTMHOT)>",
            "<NMEA(PQTMWARM)>",
            "<NMEA(PQTMCOLD)>",
            "<NMEA(PQTMGNSSSTART)>",
            "<NMEA(PQTMGNSSSTOP)>",
            "<NMEA(PQTMDEBUGON)>",
            "<NMEA(PQTMDEBUGOFF)>",
            "<NMEA(PQTMUNIQID)>",
            "<NMEA(PQTMRESETODO)>",
            "<NMEA(PQTMRESTOREPAR)>",
            "<NMEA(PQTMSAVEPAR)>",
            "<NMEA(PQTMCFGCNST, status=W, gps=1, glonass=1, galileo=1, beidou=1, qzss=0, navic=0)>",
            "<NMEA(PQTMCFGFIXRATE, status=W, fixinterval=1000)>",
            "<NMEA(PQTMSRR)>",
            "<NMEA(PQTMCFGRSID, status=W, rsid=1024)>",
            "<NMEA(PQTMCFGPROT, status=W, porttype=1, portid=1, inputprot=00000005, outputprot=00000005)>",
            "<NMEA(PQTMCFGNMEADP, status=W, utcdp=3, posdp=8, altdp=3, dopdp=2, spddp=3, cogdp=2)>",
            "<NMEA(PQTMCFGODO, status=W, state=1, initdist=10.5)>",
            "<NMEA(PQTMCFGPPS, status=W, ppsindex=1, enable=1, duration=100, ppsmode=1, polarity=1, reserved=0)>",
            "<NMEA(PQTMCFGRCVRMODE, status=W, rcvrmode=2)>",
            "<NMEA(PQTMCFGRTCM, status=W, msmtype=4, msmmode=0, msmelevthd=-90, reserved1=07, reserved2=06, ephmode=1, ephinterval=0)>",
            "<NMEA(PQTMCFGRTK, status=W, diffmode=1, relmode=1)>",
            "<NMEA(PQTMCFGSAT, status=W, systemid=1, signalid=1, masklow=FFFFFFFF)>",
            "<NMEA(PQTMCFGSAT, status=W, systemid=4, signalid=1, masklow=BFFCBFFF, maskhigh=1C003FFF)>",
            "<NMEA(PQTMCFGSIGNAL, status=W, gpssig=7, glonasssig=3, galileosig=F, beidousig=3F, qzsssig=7, navicsig=1)>",
            "<NMEA(PQTMCFGSVIN, status=W, svinmode=1, cfgcnt=3600, acclimit=1.2, ecefx=-2519265.0514, ecefy=4849534.9045, ecefz=3277834.6432)>",
            "<NMEA(PQTMCFGGEOFENCE, status=W, geofenceindex=0, geofencemode=1, reserved=0, shape=0, lat0=31.451248, lon0=117.451245, radiuslat1=100.5)>",
            "<NMEA(PQTMCFGGEOFENCE, status=W, geofenceindex=0, geofencemode=1, reserved=0, shape=3, lat0=31.451248, lon0=117.451245, radiuslat1=100.5, lon1=12.36, lat2=85.24, lon2=118.72, lat3=56.45, lon3=140.13)>",
            "<NMEA(PQTMCFGPPS, status=R, ppsindex=1)>",
            "<NMEA(PQTMCFGPPS, status=W, ppsindex=1, enable=0)>",
            "<NMEA(PQTMCFGPROT, status=R, porttype=1, portid=1)>",
            "<NMEA(PQTMCFGNMEADP, status=R)>",
            "<NMEA(PQTMCFGMSGRATE, status=R, msgname=GGA)>",
            "<NMEA(PQTMCFGGEOFENCE, status=W, geofenceindex=0, geofencemode=0)>",
            "<NMEA(PQTMCFGGEOFENCE, status=R, geofenceindex=0)>",
            "<NMEA(PQTMCFGSVIN, status=R)>",
            "<NMEA(PQTMCFGRCVRMODE, status=R)>",
            "<NMEA(PQTMCFGFIXRATE, status=R)>",
            "<NMEA(PQTMCFGRTK, status=R)>",
            "<NMEA(PQTMCFGCNST, status=R)>",
            "<NMEA(PQTMCFGODO, status=R)>",
            "<NMEA(PQTMCFGSIGNAL, status=R)>",
            "<NMEA(PQTMCFGSAT, status=R, systemid=1, signalid=1)>",
            "<NMEA(PQTMCFGRSID, status=R)>",
            "<NMEA(PQTMCFGSTATICHOLD, status=R)>",
        ]
        EXPECTED_BINARY = [
            "$PQTMCFGPPS,R,0*6B\r\n",
            "$PQTMCFGPPS,OK,0,0,0,0,0,0*21\r\n",
            "$PQTMCFGUART,W,115200*18\r\n",
            "$PQTMCFGUART,W,1,115200*05\r\n",
            "$PQTMCFGUART,W,115200,8,0,1,0*11\r\n",
            "$PQTMCFGUART,W,1,115200,8,0,0,0*0D\r\n",
            "$PQTMCFGUART,R*36\r\n",
            "$PQTMCFGUART,R,1*2B\r\n",
            "$PQTMCFGMSGRATE,W,GLL,1*0C\r\n",
            "$PQTMCFGMSGRATE,W,GLL,1,2*12\r\n",
            "$PQTMCFGMSGRATE,R,GLL*14\r\n",
            "$PQTMCFGMSGRATE,R,PQTMEPE,2*05\r\n",
            "$PQTMVERNO*58\r\n",
            "$PQTMVER,1,MODULE,LG290P03AANR01A03S,2024/04/30,10:53:07*32\r\n",
            "$PQTMHOT*4B\r\n",
            "$PQTMWARM*11\r\n",
            "$PQTMCOLD*1C\r\n",
            "$PQTMGNSSSTART*51\r\n",
            "$PQTMGNSSSTOP*09\r\n",
            "$PQTMDEBUGON*48\r\n",
            "$PQTMDEBUGOFF*06\r\n",
            "$PQTMUNIQID*16\r\n",
            "$PQTMRESETODO*09\r\n",
            "$PQTMRESTOREPAR*13\r\n",
            "$PQTMSAVEPAR*5A\r\n",
            "$PQTMCFGCNST,W,1,1,1,1,0,0*2B\r\n",
            "$PQTMCFGFIXRATE,W,1000*59\r\n",
            "$PQTMSRR*4B\r\n",
            "$PQTMCFGRSID,W,1024*06\r\n",
            "$PQTMCFGPROT,W,1,1,00000005,00000005*38\r\n",
            "$PQTMCFGNMEADP,W,3,8,3,2,3,2*39\r\n",
            "$PQTMCFGODO,W,1,10.5*4E\r\n",
            "$PQTMCFGPPS,W,1,1,100,1,1,0*73\r\n",
            "$PQTMCFGRCVRMODE,W,2*29\r\n",
            "$PQTMCFGRTCM,W,4,0,-90,07,06,1,0*25\r\n",
            "$PQTMCFGRTK,W,1,1*6C\r\n",
            "$PQTMCFGSAT,W,1,1,FFFFFFFF*4B\r\n",
            "$PQTMCFGSAT,W,4,1,BFFCBFFF,1C003FFF*60\r\n",
            "$PQTMCFGSIGNAL,W,7,3,F,3F,7,1*0E\r\n",
            "$PQTMCFGSVIN,W,1,3600,1.2,-2519265.0514,4849534.9045,3277834.6432*01\r\n",
            "$PQTMCFGGEOFENCE,W,0,1,0,0,31.451248,117.451245,100.5*18\r\n",
            "$PQTMCFGGEOFENCE,W,0,1,0,3,31.451248,117.451245,100.5,12.36,85.24,118.72,56.45,140.13*1C\r\n",
            "$PQTMCFGPPS,R,1*6A\r\n",
            "$PQTMCFGPPS,W,1,0*73\r\n",
            "$PQTMCFGPROT,R,1,1*3D\r\n",
            "$PQTMCFGNMEADP,R*37\r\n",
            "$PQTMCFGMSGRATE,R,GGA*12\r\n",
            "$PQTMCFGGEOFENCE,W,0,0*27\r\n",
            "$PQTMCFGGEOFENCE,R,0*3E\r\n",
            "$PQTMCFGSVIN,R*26\r\n",
            "$PQTMCFGRCVRMODE,R*32\r\n",
            "$PQTMCFGFIXRATE,R*71\r\n",
            "$PQTMCFGRTK,R*69\r\n",
            "$PQTMCFGCNST,R*2E\r\n",
            "$PQTMCFGODO,R*60\r\n",
            "$PQTMCFGSIGNAL,R*3A\r\n",
            "$PQTMCFGSAT,R,1,1*62\r\n",
            "$PQTMCFGRSID,R*28\r\n",
        ]
        msgs = [
            NMEAMessage("P", "QTMCFGPPS", POLL),
            NMEAMessage("P", "QTMCFGPPS", GET),
            NMEAMessage("P", "QTMCFGUART", SET, baudrate=115200),
            NMEAMessage("P", "QTMCFGUART", SET, portid=1, baudrate=115200),
            NMEAMessage("P", "QTMCFGUART", SET, baudrate=115200, databit=8, stopbit=1),
            NMEAMessage("P", "QTMCFGUART", SET, portid=1, baudrate=115200, databit=8),
            NMEAMessage("P", "QTMCFGUART", POLL),
            NMEAMessage("P", "QTMCFGUART", POLL, portid=1),
            NMEAMessage("P", "QTMCFGMSGRATE", SET, msgname="GLL", rate=1),
            NMEAMessage("P", "QTMCFGMSGRATE", SET, msgname="GLL", rate=1, msgver=2),
            NMEAMessage("P", "QTMCFGMSGRATE", POLL, msgname="GLL"),
            NMEAMessage("P", "QTMCFGMSGRATE", POLL, msgname="PQTMEPE", msgver=2),
            NMEAMessage("P", "QTMVERNO", POLL),
            NMEAMessage(
                "P",
                "QTMVER",
                GET,
                msgver=1,
                vername="MODULE",
                verstr="LG290P03AANR01A03S",
                builddate="2024/04/30",
                buildtime="10:53:07",
            ),
            NMEAMessage("P", "QTMHOT", SET),
            NMEAMessage("P", "QTMWARM", SET),
            NMEAMessage("P", "QTMCOLD", SET),
            NMEAMessage("P", "QTMGNSSSTART", SET),
            NMEAMessage("P", "QTMGNSSSTOP", SET),
            NMEAMessage("P", "QTMDEBUGON", SET),
            NMEAMessage("P", "QTMDEBUGOFF", SET),
            NMEAMessage("P", "QTMUNIQID", SET),
            NMEAMessage("P", "QTMRESETODO", SET),
            NMEAMessage("P", "QTMRESTOREPAR", SET),
            NMEAMessage("P", "QTMSAVEPAR", SET),
            NMEAMessage(
                "P",
                "QTMCFGCNST",
                SET,
                gps=1,
                glonass=1,
                galileo=1,
                beidou=1,
                qzss=0,
                navic=0,
            ),
            NMEAMessage("P", "QTMCFGFIXRATE", SET, fixinterval=1000),
            NMEAMessage("P", "QTMSRR", SET),
            NMEAMessage("P", "QTMCFGRSID", SET, rsid=1024),
            NMEAMessage(
                "P",
                "QTMCFGPROT",
                SET,
                porttype=1,
                portid=1,
                inputprot=hex2str(0x05, 8),  # hex as padded string
                outputprot=hex2str(0x05, 8),  # hex as padded string
            ),
            NMEAMessage(
                "P",
                "QTMCFGNMEADP",
                SET,
                utcdp=3,
                posdp=8,
                altdp=3,
                dopdp=2,
                spddp=3,
                cogdp=2,
            ),
            NMEAMessage(
                "P",
                "QTMCFGODO",
                SET,
                state=1,
                initdist=10.5,
            ),
            NMEAMessage(
                "P",
                "QTMCFGPPS",
                SET,
                ppsindex=1,
                enable=1,
                duration=100,
                ppsmode=1,
                polarity=1,
            ),
            NMEAMessage("P", "QTMCFGRCVRMODE", SET, rcvrmode=2),
            NMEAMessage(
                "P",
                "QTMCFGRTCM",
                SET,
                msmtype=4,
                msmmode=0,
                msmelevthd=-90,
                reserved1="07",
                reserved2="06",
                ephmode=1,
                ephinterval=0,
            ),
            NMEAMessage("P", "QTMCFGRTK", SET, diffmode=1, relmode=1),
            NMEAMessage(
                "P",
                "QTMCFGSAT",
                SET,
                systemid=1,
                signalid=1,
                masklow=hex2str(0xFFFFFFFF, 8),  # hex as padded string
            ),
            NMEAMessage(
                "P",
                "QTMCFGSAT",
                SET,
                systemid=4,
                signalid=1,
                masklow=hex2str(0xBFFCBFFF, 8),  # hex as padded string
                maskhigh=hex2str(0x1C003FFF, 8),  # hex as padded string
            ),
            NMEAMessage(
                "P",
                "QTMCFGSIGNAL",
                SET,
                gpssig=hex2str(0x07),  # hex as unpadded string
                glonasssig=hex2str(0x03),
                galileosig=hex2str(0x0F),
                beidousig=hex2str(0x3F),
                qzsssig=hex2str(0x07),
                navicsig=hex2str(0x01),
            ),
            NMEAMessage(
                "P",
                "QTMCFGSVIN",
                SET,
                svinmode=1,
                cfgcnt=3600,
                acclimit=1.2,
                ecefx=-2519265.0514,
                ecefy=4849534.9045,
                ecefz=3277834.6432,
            ),
            NMEAMessage(
                "P",
                "QTMCFGGEOFENCE",
                SET,
                geofenceindex=0,
                geofencemode=1,
                shape=0,
                lat0=31.451248,
                lon0=117.451245,
                radiuslat1=100.5,
            ),
            NMEAMessage(
                "P",
                "QTMCFGGEOFENCE",
                SET,
                geofenceindex=0,
                geofencemode=1,
                shape=3,
                lat0=31.451248,
                lon0=117.451245,
                radiuslat1=100.5,
                lon1=12.36,
                lat2=85.24,
                lon2=118.72,
                lat3=56.45,
                lon3=140.13,
            ),
            NMEAMessage(
                "P",
                "QTMCFGPPS",
                POLL,
                ppsindex=1,
            ),
            NMEAMessage(
                "P",
                "QTMCFGPPS",
                SET,
                ppsindex=1,
                enable=0,
            ),
            NMEAMessage(
                "P",
                "QTMCFGPROT",
                POLL,
                porttype=1,
                portid=1,
            ),
            NMEAMessage(
                "P",
                "QTMCFGNMEADP",
                POLL,
            ),
            NMEAMessage("P", "QTMCFGMSGRATE", POLL, msgname="GGA"),
            NMEAMessage(
                "P",
                "QTMCFGGEOFENCE",
                SET,
                geofenceindex=0,
                geofencemode=0,
            ),
            NMEAMessage(
                "P",
                "QTMCFGGEOFENCE",
                POLL,
                geofenceindex=0,
            ),
            NMEAMessage(
                "P",
                "QTMCFGSVIN",
                POLL,
            ),
            NMEAMessage("P", "QTMCFGRCVRMODE", POLL),
            NMEAMessage("P", "QTMCFGFIXRATE", POLL),
            NMEAMessage("P", "QTMCFGRTK", POLL),
            NMEAMessage("P", "QTMCFGCNST", POLL),
            NMEAMessage("P", "QTMCFGODO", POLL),
            NMEAMessage("P", "QTMCFGSIGNAL", POLL),
            NMEAMessage("P", "QTMCFGSAT", POLL, systemid=1, signalid=1),
            NMEAMessage("P", "QTMCFGRSID", POLL),
        ]
        for i, msg in enumerate(msgs):
            self.assertEqual(str(msg), EXPECTED_PARSED[i])
            self.assertEqual(msg.serialize().decode(), EXPECTED_BINARY[i])

    def testNMEAQUECTELAIRGET(self):  # Test Quectel PAIR GET responses
        EXPECTED_RESULTS = (
            "<NMEA(PAIR001, commandid=4, result=0)>",
            "<NMEA(PAIR001, commandid=3, result=1)>",
            "<NMEA(PAIR001, commandid=3, result=0)>",
            "<NMEA(PAIR001, commandid=4, result=0)>",
            "<NMEA(PAIR001, commandid=5, result=0)>",
            "<NMEA(PAIR001, commandid=6, result=0)>",
            "<NMEA(PAIR001, commandid=7, result=0)>",
            "<NMEA(PAIR010, type=0, subsystem=0, wno=2044, tow=369413)>",
            "<NMEA(PAIR001, commandid=50, result=0)>",
            "<NMEA(PAIR001, commandid=51, result=0)>",
            "<NMEA(PAIR051, time=1000)>",
            "<NMEA(PAIR001, commandid=58, result=0)>",
            "<NMEA(PAIR001, commandid=59, result=0)>",
            "<NMEA(PAIR059, minsnr=15)>",
            "<NMEA(PAIR001, commandid=62, result=0)>",
            "<NMEA(PAIR001, commandid=63, result=0)>",
            "<NMEA(PAIR063, type=0, rate=3)>",
            "<NMEA(PAIR001, commandid=66, result=0)>",
            "<NMEA(PAIR001, commandid=67, result=0)>",
            "<NMEA(PAIR067, gpsEnabled=1, glonassEnabled=1, galileoEnabled=1, bdsEnabled=1, qzssEnabled=1, navicEnabled=0)>",
            "<NMEA(PAIR001, commandid=70, result=0)>",
            "<NMEA(PAIR001, commandid=71, result=0)>",
            "<NMEA(PAIR071, speedThreshold=0.4)>",
            "<NMEA(PAIR001, commandid=72, result=0)>",
            "<NMEA(PAIR001, commandid=73, result=0)>",
            "<NMEA(PAIR073, degree=5)>",
            "<NMEA(PAIR001, commandid=74, result=0)>",
            "<NMEA(PAIR001, commandid=75, result=0)>",
            "<NMEA(PAIR075, enabled=1)>",
            "<NMEA(PAIR001, commandid=80, result=0)>",
            "<NMEA(PAIR001, commandid=81, result=0)>",
            "<NMEA(PAIR081, navMode=0)>",
            "<NMEA(PAIR001, commandid=86, result=0)>",
            "<NMEA(PAIR001, commandid=87, result=0)>",
            "<NMEA(PAIR087, status=0)>",
            "<NMEA(PAIR001, commandid=100, result=0)>",
            "<NMEA(PAIR001, commandid=101, result=0)>",
            "<NMEA(PAIR101, nmeaMode=1, reserved=0)>",
            "<NMEA(PAIR001, commandid=382, result=0)>",
            "<NMEA(PAIR001, commandid=3, result=0)>",
            "<NMEA(PAIR001, commandid=104, result=0)>",
            "<NMEA(PAIR001, commandid=2, result=0)>",
            "<NMEA(PAIR001, commandid=105, result=0)>",
            "<NMEA(PAIR105, dualBandEnabled=1)>",
            "<NMEA(PAIR001, commandid=382, result=0)>",
            "<NMEA(PAIR001, commandid=391, result=0)>",
            "<NMEA(PAIRSPF, status=0)>",
            "<NMEA(PAIRSPF5, status=0)>",
            "<NMEA(PAIRSPF, status=1)>",
            "<NMEA(PAIRSPF5, status=1)>",
            "<NMEA(PAIRSPF, status=2)>",
            "<NMEA(PAIRSPF5, status=2)>",
            "<NMEA(PAIRSPF, status=3)>",
            "<NMEA(PAIRSPF5, status=3)>",
            "<NMEA(PAIR001, commandid=400, result=0)>",
            "<NMEA(PAIR001, commandid=401, result=0)>",
            "<NMEA(PAIR401, mode=2)>",
            "<NMEA(PAIR001, commandid=410, result=0)>",
            "<NMEA(PAIR001, commandid=411, result=0)>",
            "<NMEA(PAIR411, enabled=1)>",
            "<NMEA(PAIR001, commandid=420, result=0)>",
            "<NMEA(PAIR001, commandid=421, result=0)>",
            "<NMEA(PAIR421, enabled=1)>",
            "<NMEA(PAIR001, commandid=433, result=0)>",
            "<NMEA(PAIR433, mode=-1)>",
            "<NMEA(PAIR001, commandid=434, result=0)>",
            "<NMEA(PAIR001, commandid=435, result=0)>",
            "<NMEA(PAIR435, enabled=1)>",
            "<NMEA(PAIR001, commandid=436, result=0)>",
            "<NMEA(PAIR001, commandid=437, result=0)>",
            "<NMEA(PAIR437, enabled=1)>",
            "<NMEA(PAIR001, commandid=650, result=0)>",
            "<NMEA(PAIR650, second=0)>",
            "<NMEA(PAIR001, commandid=752, result=0)>",
            "<NMEA(PAIR001, commandid=864, result=0)>",
            "<NMEA(PAIR001, commandid=865, result=0)>",
            "<NMEA(PAIR865, baudrate=115200)>",
            "<NMEA(PAIR001, commandid=866, result=0)>",
            "<NMEA(PAIR001, commandid=867, result=0)>",
            "<NMEA(PAIR867, flowControl=0)>",
            "<NMEA(PAIR6011, type=1, rate=0)>",
        )

        i = 0
        raw = 0
        with open(
            os.path.join(DIRNAME, "pygpsdata_quectel_pair_get.log"), "rb"
        ) as stream:
            nmr = NMEAReader(stream)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, len(EXPECTED_RESULTS))

    def testNMEAQUECTELAIRSET(self):  # Test Quectel PAIR SET commands
        EXPECTED_RESULTS = (
            "<NMEA(PAIR002)>",
            "<NMEA(PAIR003)>",
            "<NMEA(PAIR004)>",
            "<NMEA(PAIR005)>",
            "<NMEA(PAIR007)>",
            "<NMEA(PAIR050, time=1000)>",
            "<NMEA(PAIR058, minsnr=15)>",
            "<NMEA(PAIR062, type=0, rate=3)>",
            "<NMEA(PAIR066, gpsEnabled=1, glonassEnabled=0, galileoEnabled=1, bdsEnabled=1, qzssEnabled=0, navicEnabled=0)>",
            "<NMEA(PAIR070, speedThreshold=4)>",
            "<NMEA(PAIR072, degree=5)>",
            "<NMEA(PAIR074, enabled=1)>",
            "<NMEA(PAIR080, navMode=1)>",
            "<NMEA(PAIR081)>",
            "<NMEA(PAIR086, status=1)>",
            "<NMEA(PAIR382, enabled=1)>",
            "<NMEA(PAIR003)>",
            "<NMEA(PAIR104, dualBandEnabled=0)>",
            "<NMEA(PAIR002)>",
            "<NMEA(PAIR382, enabled=1)>",
            "<NMEA(PAIR391, cmdType=1)>",
            "<NMEA(PAIR400, mode=2)>",
            "<NMEA(PAIR410, enabled=1)>",
            "<NMEA(PAIR420, enabled=1)>",
            "<NMEA(PAIR432, mode=1)>",
            "<NMEA(PAIR434, enabled=1)>",
            "<NMEA(PAIR436, enabled=1)>",
            "<NMEA(PAIR511)>",
            "<NMEA(PAIR382, enabled=1)>",
            "<NMEA(PAIR003)>",
            "<NMEA(PAIR511)>",
            "<NMEA(PAIR002)>",
            "<NMEA(PAIR513)>",
            "<NMEA(PAIR382, enabled=1)>",
            "<NMEA(PAIR003)>",
            "<NMEA(PAIR513)>",
            "<NMEA(PAIR002)>",
            "<NMEA(PAIR650, second=0)>",
            "<NMEA(PAIR752, ppsType=2)>",
            "<NMEA(PAIR864, portType=0, portIndex=0, baudrate=115200)>",
            "<NMEA(PAIR866, portType=0, portIndex=0, flowControl=1)>",
            "<NMEA(PAIR6010, type=0, rate=1)>",
        )

        i = 0
        raw = 0
        with open(
            os.path.join(DIRNAME, "pygpsdata_quectel_pair_set.log"), "rb"
        ) as stream:
            nmr = NMEAReader(stream, msgmode=SET)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, len(EXPECTED_RESULTS))

    def testNMEAQUECTELAIRPOLL(self):  # Test Quectel PAIR POLL commands
        EXPECTED_RESULTS = (
            "<NMEA(PAIR051)>",
            "<NMEA(PAIR059)>",
            "<NMEA(PAIR063, type=0)>",
            "<NMEA(PAIR067)>",
            "<NMEA(PAIR071)>",
            "<NMEA(PAIR073)>",
            "<NMEA(PAIR075)>",
            "<NMEA(PAIR081)>",
            "<NMEA(PAIR087)>",
            "<NMEA(PAIR101)>",
            "<NMEA(PAIR105)>",
            "<NMEA(PAIR401)>",
            "<NMEA(PAIR411)>",
            "<NMEA(PAIR421)>",
            "<NMEA(PAIR433)>",
            "<NMEA(PAIR435)>",
            "<NMEA(PAIR437)>",
            "<NMEA(PAIR865, portType=0, portIndex=0)>",
            "<NMEA(PAIR867, portType=0, portIndex=0)>",
            "<NMEA(PAIR6011, type=1)>",
        )

        i = 0
        raw = 0
        with open(
            os.path.join(DIRNAME, "pygpsdata_quectel_pair_poll.log"), "rb"
        ) as stream:
            nmr = NMEAReader(stream, msgmode=POLL)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, len(EXPECTED_RESULTS))



"""
NMEA Protocol Proprietary Output (GET) payload definitions

THESE ARE THE PAYLOAD DEFINITIONS FOR PROPRIETARY _GET_ MESSAGES _FROM_
THE RECEIVER (e.g. Periodic Navigation Data; Poll Responses; Info messages).

NB: Attribute names must be unique within each message id.
NB: Avoid reserved names 'msgID', 'talker', 'payload', 'checksum'.

NB: Repeating groups must be defined as a tuple thus
    'group': ('numr', {dict})
    where
    - 'numr' is either:
       a) an integer representing a fixed number of repeats e.g 32
       b) a string representing the name of a preceding attribute
          containing the number of repeats e.g. 'numCh'
       c) 'None' for an indeterminate repeating group
          (only one such group is permitted per message type)
    - {dict} is the nested dictionary containing the repeating
      attributes

Created on 4 Mar Sep 2021

While the NMEA 0183 © protocol is proprietary, the information here
has been collated from public domain sources.

:author: semuadmin (Steve Smith)
"""

# pylint: disable=too-many-lines, duplicate-code

from pynmeagps.nmeatypes_core import (
    CH,
    DE,
    DM,
    DT,
    HX,
    IN,
    LA,
    LAD,
    LN,
    LND,
    QS,
    ST,
    TM,
)

NMEA_PAYLOADS_GET_PROP = {
    # *********************************************
    # JVC KENWOOD PROPRIETARY MESSAGES
    # *********************************************
    "KLDS": {
        "time": TM,
        "status": CH,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "sog": DE,
        "cog": DE,
        "dat": DT,
        "declination": DE,
        "dec_dir": CH,
        "fleet": DE,
        "senderid": ST,
        "senderstatus": DE,
        "reserved": DE,
    },
    "KLSH": {
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "time": TM,
        "status": CH,
        "fleetId": ST,
        "deviceId": ST,
    },
    "KNDS": {
        "time": TM,
        "status": CH,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "sog": DE,
        "cog": DE,
        "date": DT,
        "declination": DE,
        "dec_dir": CH,
        "senderid": ST,
        "senderstatus": DE,
        "reserved": DE,
    },
    "KNSH": {
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "time": TM,
        "status": CH,
        "senderid": ST,
    },
    "KWDWPL": {
        "time": TM,
        "status": CH,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "sog": DE,
        "cog": DE,
        "date": DT,
        "alt": DE,
        "wpt": ST,
        "ts": ST,
    },
    # *********************************************
    # MAGELLAN PROPRIETARY MESSAGES
    # *********************************************
    "MGNWPL": {
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "alt": DE,
        "alt_unit": CH,
        "wpt": ST,
        "comment": ST,
        "icon": ST,
        "type": ST,
    },
    # *********************************************
    # GARMIN PROPRIETARY MESSAGES
    # *********************************************
    "GRME": {  # estimated error information
        "HPE": DE,
        "HPEUnit": CH,
        "VPE": DE,
        "VPEUnit": CH,
        "EPE": DE,
        "EPEUnit": CH,
    },
    "GRMF": {  # GPS fix data sentence
        "week": IN,
        "secs": IN,
        "date": DT,
        "time": TM,
        "leapsec": IN,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "mode": CH,
        "fix": IN,
        "spd": DE,
        "course": IN,
        "PDOP": DE,
        "TDOP": DE,
    },
    "GRMH": {  # aviation height and VNAV data
        "status": CH,
        "vspd": DE,
        "verr": DE,
        "spdtgt": DE,
        "spdwpt": DE,
        "height": DE,
        "trk": DE,
        "course": DE,
    },
    "GRMM": {  # map datum
        "dtm": ST,
    },
    "GRMT": {  # sensor status information
        "ver": ST,
        "ROMtest": CH,
        "rcvrtest": CH,
        "stortest": CH,
        "rtctest": CH,
        "osctest": CH,
        "datatest": CH,
        "temp": DE,
        "cfgdata": CH,
    },
    "GRMV": {  # 3D velocity information
        "velE": DE,
        "velN": DE,
        "velZ": DE,
    },
    "GRMZ": {  # altitude
        "alt": DE,
        "altUnit": CH,
        "fix": IN,
    },
    "GRMB": {  # DGPS Beacon information
        "freq": DE,
        "bps": IN,
        "snr": IN,
        "quality": IN,
        "dist": DE,
        "status": IN,
        "fixsrc": CH,
        "mode": CH,
    },
    # *********************************************
    # U-BLOX PROPRIETARY MESSAGES
    # *********************************************
    "UBX00": {
        "msgId": ST,  # '00'
        "time": TM,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "altRef": DE,
        "navStat": ST,
        "hAcc": DE,
        "vAcc": DE,
        "SOG": DE,
        "COG": DE,
        "vVel": DE,
        "diffAge": DE,
        "HDOP": DE,
        "VDOP": DE,
        "TDOP": DE,
        "numSVs": IN,
        "reserved": IN,
        "DR": IN,
    },
    "UBX03": {
        "msgId": ST,  # '03'
        "numSv": IN,
        "groupSV": (
            "numSv",
            {  # repeating group * numSv
                "svid": IN,
                "status": CH,
                "azi": DE,
                "ele": DE,
                "cno": IN,
                "lck": IN,
            },
        ),
    },
    "UBX04": {
        "msgId": ST,  # '04'
        "time": TM,
        "date": DT,
        "utcTow": ST,
        "utcWk": ST,
        "leapSec": ST,
        "clkBias": DE,
        "clkDrift": DE,
        "tpGran": IN,
    },
    "UBX05": {  # deprecated, for backwards compat only
        "msgId": ST,  # '05'
        "pulses": IN,
        "period": IN,
        "gyroMean": IN,
        "temperature": DE,
        "direction": CH,
        "pulseScaleCS": IN,
        "gyroScaleCS": IN,
        "gyroBiasCS": IN,
        "pulseScale": DE,
        "gyroBias": DE,
        "gyroScale": DE,
        "pulseScaleAcc": IN,
        "gyroBiasAcc": IN,
        "gyroScaleAcc": IN,
        "measUsed": HX,
    },
    # ***************************************************************
    # Trimble Proprietary message types
    # https://receiverhelp.trimble.com/alloy-gnss/en-us/NMEA-0183messages_MessageOverview.html
    # ***************************************************************
    "ASHR": {
        "utctime": TM,
        "trueHdg": DE,
        "trueHdgInd": CH,  # "T"
        "roll": DE,
        "pitch": DE,
        "reserved": ST,
        "rollAcc": DE,
        "pitchAcc": DE,
        "hdgAcc": DE,
        "gnssQual": IN,
        "imuAlign": IN,
    },
    "ASHRALR": {
        "msgId": ST,  # "ALR"
        "alarmCode": IN,
        "alarmSubcode": IN,
        "streamId": CH,
        "alarmCat": ST,
        "alarmLevel": IN,
        "desc": ST,
    },
    "ASHRARA": {
        "msgId": ST,  # "ARA"
        "valid": CH,  # = "0" when valid
        "utctime": TM,
        "hdgSpeed": DE,
        "pitchSpeed": DE,
        "rollSpeed": DE,
        "hdgAcc": DE,
        "pitchAcc": DE,
        "rollAcc": DE,
        "reserved": ST,
    },
    "ASHRARR": {
        "msgId": ST,  # "ARR"
        "vectNum": IN,
        "vectMode": IN,
        "sip": IN,
        "utctime": TM,
        "antEcefX": DE,
        "antEcefY": DE,
        "antEcefZ": DE,
        "coord1std": DE,
        "coord2std": DE,
        "coord3std": DE,
        "coord12corr": DE,
        "coord13corr": DE,
        "coord23corr": DE,
        "refId": CH,
        "vectFrame": IN,  # 0 = XYZ
        "vectOpt": IN,
        "clkAssum": IN,
    },
    "ASHRATT": {
        "msgId": ST,  # "ATT"
        "weekTime": DE,
        "trueHdg": DE,
        "pitch": DE,
        "roll": DE,
        "carrierRmsErr": DE,
        "baselineRmsErr": DE,
        "ambiguity": IN,
    },
    "ASHRBTS": {
        "msgId": ST,  # "BTS"
        "portgroup": (
            3,
            {
                "port": CH,  # C, H or T
                "connected": IN,
                "name": ST,
                "addr": ST,
                "linkQual": IN,
            },
        ),
    },
    "ASHRCAP": {
        "msgId": ST,  # "CAP"
        "name": ST,
        "L1Noffset": DE,
        "L1Eoffset": DE,
        "L1Voffset": DE,
        "L2Noffset": DE,
        "L2Eoffset": DE,
        "L2Voffset": DE,
    },
    "ASHRCPA": {
        "msgId": ST,  # "CPA"
        "antHeight": DE,
        "antRadius": DE,
        "vertOffset": DE,
        "horAzi": DE,
        "horDist": DE,
        "reserved1": ST,
        "reserved2": ST,
        "reserved3": ST,
        "reserved4": ST,
    },
    "ASHRCPO": {
        "msgId": ST,  # "CPO"
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "height": DE,
    },
    "ASHRDDM": {
        "msgId": ST,  # "DDM"
        "corrPort": CH,
        "msgTransport": ST,
        "msgIdent": ST,
        "count": IN,
        "baseId": ST,
        "timeLag": DE,
        "corrAge": DE,
        "attr": ST,
    },
    "ASHRDDS": {
        "msgId": ST,  # "DDS"
        "diffDecodeNum": IN,
        "timeTag": TM,
        "numMsgDecoded": IN,
        "port": CH,
        "prot": ST,
        "timeWindow": IN,
        "percLinkQual": IN,
        "percDeselectInfo": IN,
        "percCrc": IN,
        "latencyStd": IN,
        "latencyMean": IN,
        "epochIntMean": IN,
        "epochIntMin": IN,
        "numMsgDetected": IN,
        "msgType": ST,
        "lastMsgInt": DE,
        "lastMsgAge": DE,
    },
    "ASHRHPR": {
        "msgId": ST,  # "HPR"
        "utctime": TM,
        "trueHdg": DE,
        "pitch": DE,
        "roll": DE,
        "carrierRmsErr": DE,
        "baselineRmsErr": DE,
        "ambiguity": IN,
        "attStatus": IN,
        "antBaseline": ST,
        "PDOP": DE,
    },
    "ASHRLTN": {
        "msgId": ST,  # "LTN"
        "latency": DE,
    },
    "ASHRMDM": {
        "msgId": ST,  # "MDM"
        "port": ST,
        "baud": IN,
        "state": ST,
        "powerMode": ST,
        "pinCode": ST,
        "protocol": IN,
        "csdMode": IN,  # not used
        "accessPoint": ST,
        "login": ST,
        "password": ST,
        "phoneNum": ST,
        "autoDial": CH,
        "maxRedial": IN,
        "model": ST,
        "selMode": IN,
        "gsmAnt": ST,
    },
    "ASHRPBN": {
        "msgId": ST,  # "PBN",
        "gpsWeek": DE,
        "siteName": ST,
        "navX": DE,
        "navY": DE,
        "navT": DE,
        "navXdot": DE,
        "navYdot": DE,
        "navTdot": DE,
        "PDOP": DE,
    },
    "ASHRPOS": {
        "msgId": ST,  # "POS"
        "solnType": IN,
        "sip": IN,
        "utctime": TM,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "alt": DE,
        "corrAge": DE,
        "trueTrk": DE,
        "sog": DE,
        "vertVel": DE,
        "PDOP": DE,
        "HDOP": DE,
        "VDOP": DE,
        "TDOP": DE,
        "baseId": ST,
    },
    "ASHRPTT": {
        "msgId": ST,  # "PTT"
        "dow": IN,
        "timeTag": ST,
    },
    "ASHRPWR": {
        "msgId": ST,  # "PWR"
        "powerSource": IN,
        "voltsOut": DE,
        "reserved1": ST,
        "percRemaining": IN,
        "reserved2": ST,
        "voltsIn": DE,
        "chargeStatus": IN,
        "reserved3": ST,
        "tempInt": DE,
        "tempBatt": DE,
    },
    "ASHRRCS": {
        "msgId": ST,  # "RCS"
        "recordStatus": CH,
        "memory": IN,
        "fileName": ST,
        "recordRate": IN,
        "occupationType": IN,
        "occupationState": IN,
        "occupationName": ST,
    },
    "ASHRSBD": {
        "msgId": ST,  # "SBD"
        "siv": IN,
        "satgroup": (  # repeating group * siv
            "siv",
            {
                "prn": IN,
                "azi": DE,
                "ele": DE,
                "snrB1": DE,
                "snrB2": DE,
                "snrB3": DE,
                "usageStatus": ST,
                "corrStatus": ST,
            },
        ),
    },
    "ASHRSGA": {
        "msgId": ST,  # "SGA"
        "siv": IN,
        "satgroup": (  # repeating group * siv
            "siv",
            {
                "prn": IN,
                "azi": DE,
                "ele": DE,
                "snrE1": DE,
                "snrE5b": DE,
                "snrE5a": DE,
                "usageStatus": ST,
                "corrStatus": ST,
            },
        ),
    },
    "ASHRSGL": {
        "msgId": ST,  # "SGL"
        "siv": IN,
        "satgroup": (  # repeating group * siv
            "siv",
            {
                "prn": IN,
                "azi": DE,
                "ele": DE,
                "snrL1": DE,
                "snrL2": DE,
                "snrL3": DE,
                "usageStatus": ST,
                "corrStatus": ST,
            },
        ),
    },
    "ASHRSGO": {
        "msgId": ST,  # "SGO"
        "siv": IN,
        "satgroup": (  # repeating group * siv
            "siv",
            {
                "prn": IN,
                "azi": DE,
                "ele": DE,
                "snrE1": DE,
                "snrE5b": DE,
                "snrE5a": DE,
                "snrE6": DE,
                "reserved": ST,
                "usageStatus": ST,
                "corrStatus": ST,
            },
        ),
    },
    "ASHRSGP": {
        "msgId": ST,  # "SGP"
        "siv": IN,
        "satgroup": (  # repeating group * siv
            "siv",
            {
                "prn": IN,
                "azi": DE,
                "ele": DE,
                "snrL1": DE,
                "snrL2": DE,
                "snrL3": DE,
                "usageStatus": ST,
                "corrStatus": ST,
            },
        ),
    },
    "ASHRSIR": {
        "msgId": ST,  # "SIR"
        "siv": IN,
        "satgroup": (  # repeating group * siv
            "siv",
            {
                "prn": IN,
                "azi": DE,
                "ele": DE,
                "reserved2": ST,
                "reserved3": ST,
                "snrL5": DE,
                "usageStatus": ST,
                "corrStatus": ST,
            },
        ),
    },
    "ASHRSLB": {
        "msgId": ST,  # "SLB"
        "siv": IN,
        "satgroup": (  # repeating group * siv
            "siv",
            {
                "satNum": IN,
                "contTrkInt": DE,
                "azi": DE,
                "ele": DE,
                "snr": DE,
            },
        ),
    },
    "ASHRSQZ": {
        "msgId": ST,  # "SQZ"
        "siv": IN,
        "satgroup": (  # repeating group * siv
            "siv",
            {
                "prn": IN,
                "azi": DE,
                "ele": DE,
                "snrL1": DE,
                "snrL2": DE,
                "snrL3": DE,
                "usageStatus": ST,
                "corrStatus": ST,
            },
        ),
    },
    "ASHRSSB": {
        "msgId": ST,  # "SSB"
        "siv": IN,
        "satgroup": (  # repeating group * siv
            "siv",
            {
                "prn": IN,
                "azi": DE,
                "ele": DE,
                "snrL1": DE,
                "reserved": ST,
                "snrL5": DE,
                "usageStatus": ST,
                "corrStatus": ST,
            },
        ),
    },
    "ASHRTEM": {
        "msgId": ST,  # "TEM"
        "temp": DE,  # 1/1000 degrees
    },
    "ASHRTHS": {
        "msgId": ST,  # "THS"
        "lastComputedHdg": DE,
        "solnStatus": CH,
    },
    "ASHRTTT": {
        "msgId": ST,  # "TTT"
        "dow": IN,
        "timeTag": TM,
    },
    "ASHRVCR": {
        "msgId": ST,  # "VCR"
        "baselineNum": IN,
        "baselineMode": IN,
        "sip": IN,
        "utctime": TM,
        "antEcefX": DE,
        "antEcefY": DE,
        "antEcefZ": DE,
        "coord1std": DE,
        "coord2std": DE,
        "coord3std": DE,
        "coord12corr": DE,
        "coord13corr": DE,
        "coord23corr": DE,
        "baseId": ST,
        "baselineCoordId": ST,  # 0 = XYZ
    },
    "ASHRVCT": {
        "msgId": ST,  # "VCT"
        "baselineMode": IN,
        "sip": IN,
        "utctime": TM,
        "antEcefX": DE,
        "antEcefY": DE,
        "antEcefZ": DE,
        "coord1std": DE,
        "coord2std": DE,
        "coord3std": DE,
        "coord12corr": DE,
        "coord13corr": DE,
        "coord23corr": DE,
        "baseId": ST,
        "baselineCoordId": ST,  # 0 = XYZ
        "baselineNum": IN,
        "vrs": IN,
        "staticModeAssumption": IN,
    },
    "ASHRVEL": {
        "msgId": ST,  # "VEL"
        "velE": DE,
        "velN": DE,
        "velV": DE,
        "velERmsErr": DE,
        "velNRmsErr": DE,
        "velVRmsErr": DE,
        "velSmoothInt": IN,
    },
    "FUGDP": {
        "type": ST,  # GPS (GP), GLONASS (GL) or GNSS (GN)
        "utctime": TM,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "siv": IN,
        "dpvoaQual": ST,
        "dgnssMode": ST,
        "smajErr": DE,
        "sminErr": DE,
        "dirErr": IN,
    },
    "GPPADV110": {
        "msgId": ST,  # "110"
        "lat": DE,
        "lon": DE,
        "height": DE,
        "ele": DE,  # optional
        "azi": DE,  # optional
    },
    "GPPADV120": {
        "msgId": ST,  # "120"
        "satgrp": (  # repeating group * 2
            2,
            {
                "prn": IN,
                "ele": DE,
                "azi": DE,
            },
        ),
    },
    "TNLAVR": {
        "msgId": ST,  # "AVR"
        "utctime": TM,
        "yaw": DE,
        "yawc": ST,  # "Yaw"
        "tilt": DE,
        "tiltc": ST,  # "Tilt"
        "roll": DE,
        "rollc": ST,  # "Roll"
        "range": DE,
        "gpsQual": IN,
        "PDOP": DE,
        "sip": IN,
    },
    "TNLBPQ": {
        "msgId": ST,  # "BPQ"
        "utctime": TM,
        "utcdate": DT,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "height": ST,  # starts with "EHT-"
        "hunit": CH,  # 'M'
        "gpsQual": IN,
    },
    "TNLDG": {
        "msgId": ST,  # "DG" TODO check
        "strength": DE,
        "snr": DE,
        "freq": DE,
        "bitRate": IN,
        "chan": IN,
        "trkStatus": CH,
        "trkPerf": IN,
    },
    "TNLEVT": {
        "msgId": ST,  # "EVT"
        "utctime": TM,
        "port": IN,
        "numEvents": IN,
        "wno": IN,
        "dow": IN,
        "leaps": IN,
    },
    "TNLGGK": {
        "msgId": ST,  # "GGK"
        "utctime": TM,
        "utcdate": DT,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "gpsQual": IN,
        "sip": IN,
        "DOP": DE,
        "height": ST,  # starts with "EHT-"
        "hunit": CH,  # 'M'
    },
    "TNLGGKx": {
        "msgId": ST,  # "GGK"
        "utctime": TM,
        "utcdate": DT,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "posType": IN,
        "sip": IN,
        "PDOP": DE,
        "height": DE,
        "numExt": IN,
        "sigmaE": DE,
        "sigmaN": DE,
        "sigmaV": DE,
        "propogationAge": DE,
    },
    "TNLPJK": {
        "msgId": ST,  # "PJK"
        "utctime": TM,
        "utcdate": DT,
        "northing": DE,
        "nunit": CH,  # 'N'
        "easting": DE,
        "eunit": CH,  # 'E"
        "gpsQual": IN,
        "sip": IN,
        "DOP": DE,
        "height": ST,  # starts with "EHT-" or "GHT-"
        "hunit": CH,  # 'M'
    },
    "TNLPJT": {
        "msgId": ST,  # "PJT"
        "coordName": ST,
        "projName": ST,
    },
    "TNLREX": {
        "msgId": ST,  # "REX"
        "utctime": TM,
        "constUsed": ST,
        "sip": IN,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "alt": DE,
        "latSigmaErr": DE,
        "lonSigmaErr": DE,
        "heightSigmaErr": DE,
        "vectE": DE,
        "vectN": DE,
        "vectV": DE,
        "vectESigmaErr": DE,
        "vectNSigmaErr": DE,
        "vectVSigmaErr": DE,
        "corrSource": ST,
    },
    "TNLVGK": {
        "msgId": ST,  # "VGK"
        "utctime": TM,
        "utcdate": DM,  # mmddyy
        "vectE": DE,
        "vectN": DE,
        "vectV": DE,
        "gpsQual": IN,
        "sip": IN,
        "DOP": DE,
        "vunit": CH,  # 'M'
    },
    "TNLVHD": {
        "msgId": ST,  # "VHD"
        "utctime": TM,
        "utcdate": DM,  # mmddyy
        "azi": DE,
        "aziRate": DE,
        "ele": DE,
        "eleRate": DE,
        "range": DE,
        "rangeRate": DE,
        "gpsQual": IN,
        "sip": IN,
        "PDOP": DE,
        "unit": CH,  # 'M'
    },
    # *********************************************
    # Furuno
    # https://www.furuno.com/en/support/sdk/
    # *********************************************
    "FECGPATT": {
        "msgId": ST,  # 'GPatt'
        "yaw": DE,
        "pitch": DE,
        "roll": DE,
    },
    "FECGPHVE": {
        "msgId": ST,  # 'GPhve'
        "heave": DE,
        "status": CH,  # 'A'
    },
    # ***************************************************************
    # Locosys Proprietary Messages GET
    # https://www.locosystech.com/Templates/att/LS2303x-UDG_datasheet_v0.6.pdf?lng=en
    # ***************************************************************
    "LSR": {
        "msgType": ST,  # e.g. "SLOPE", "MEMS", "ATTIT"
        "value": IN,  # e.g. 0
        "response": ST,  # e.g. "OK"
    },
    "LSVD": {
        "velE": DE,  # cm/s
        "velN": DE,  # cm/s
        "velD": DE,  # cm/s
        "velEdev": DE,  # cm/s deviation
        "velNdev": DE,  # cm/s deviation
        "velDdev": DE,  # cm/s deviation
    },
    "INVMINR": {
        "status": CH,  # 0 uninitialized; 1、2 calibrating/initializing; 3 calibration done
    },
    "INVMSTR": {
        "value": IN,
    },
    "INVMSLOPE": {
        "slope": DE,  # degrees
        "altDiff": DE,  # meters
        "moveDist": DE,  # meters
        "slopeAccu": DE,  # degrees
        "altDiffAccu": DE,  # meters
        "moveDistAccu": DE,  # meters
    },
    "INVMIMU": {
        "timeSecond": DE,  # sec timestamp
        "accelX": DE,  # m/s^2
        "accelY": DE,  # m/s^2
        "accelZ": DE,  # m/s^2
        "gyroX": DE,  # degree/s
        "gyroY": DE,  # degree/s
        "gyroZ": DE,  # degree/s
    },
    "INVMATTIT": {
        "roll": DE,  # degree
        "pitch": DE,  # degree
        "yaw": DE,  # degree heading
    },
    # ***************************************************************
    # Quectel LC29H / LC79H Proprietary GET message types
    # https://www.quectel.com/download/quectel_lc29hlc79h_series_gnss_protocol_specification_v1-5/
    # quectel_lc29hlc79h_series_gnss_protocol_specification_v1-5.pdf
    # ***************************************************************
    "AIR001": {
        "commandid": IN,
        "result": IN,
    },
    "AIR010": {
        "type": IN,
        "subsystem": IN,
        "wno": IN,
        "tow": IN,
    },
    "AIR051": {
        "time": IN,
    },
    "AIR059": {
        "minsnr": IN,
    },
    "AIR063": {
        "type": IN,
        "rate": IN,
    },
    "AIR067": {
        "gpsEnabled": IN,
        "glonassEnabled": IN,
        "galileoEnabled": IN,
        "bdsEnabled": IN,
        "qzssEnabled": IN,
        "navicEnabled": IN,
    },
    "AIR071": {
        "speedThreshold": DE,
    },
    "AIR073": {
        "degree": IN,
    },
    "AIR075": {
        "enabled": IN,
    },
    "AIR081": {
        "navMode": IN,
    },
    "AIR087": {
        "status": IN,
    },
    "AIR101": {
        "nmeaMode": IN,
        "reserved": IN,
    },
    "AIR105": {
        "dualBandEnabled": IN,
    },
    "AIR391": {
        "cmdType": IN,
    },
    "AIR401": {
        "mode": IN,
    },
    "AIR411": {
        "enabled": IN,
    },
    "AIR421": {
        "enabled": IN,
    },
    "AIR433": {
        "mode": IN,
    },
    "AIR435": {
        "enabled": IN,
    },
    "AIR436": {
        "enabled": IN,
    },
    "AIR437": {"enabled": IN},
    "AIR650": {
        "second": IN,
    },
    "AIR865": {
        "baudrate": IN,
    },
    "AIR867": {
        "flowControl": IN,
    },
    "AIR6011": {
        "type": IN,
        "rate": IN,
    },
    "AIRSPF": {
        "status": IN,
    },
    "AIRSPF5": {
        "status": IN,
    },
    # ***************************************************************
    # Quectel LG290P Proprietary message types
    # https://quectel.com/content/uploads/2024/09/Quectel_LG290P03_GNSS_Protocol_Specification_V1.0.pdf
    #
    # status attribute must be set to:
    # 'R' for POLL messages (queries)
    # 'W' for SET messages (commands)
    # 'OK' for GET acknowledgements (ACK)
    # 'ERROR' for GET errors (NAK)
    # ***************************************************************
    "QTMACK": {
        "status": QS,  # OK
    },
    "QTMNAK": {
        "status": QS,  # ERROR
        "errcode": IN,  # 1 = invalid parms, 2 = failed exec, 3 = unsupported
    },
    "QTMCFGAIC": {
        "status": QS,
        "state": IN,
    },
    "QTMCFGCNST": {
        "status": QS,
        "gps": IN,
        "glonass": IN,
        "galileo": IN,
        "beidou": IN,
        "qzss": IN,
        "navic": IN,
    },
    "QTMCFGDR": {
        "status": QS,
        "drstate": IN,
    },
    "QTMCFGDRRTD": {"status": QS, "time": IN, "distance": IN},
    "QTMCFGEINSMSG": {
        "type": IN,
        "ins_enabled": IN,
        "imu_enabled": IN,
        "gps_enabled": IN,
        "rate": IN,
    },
    "QTMCFGFIXRATE": {
        "status": QS,
        "fixinterval": IN,
    },
    "QTMCFGGEOFENCE": {
        "status": QS,
        "geofenceindex": IN,
        "geofencemode": IN,
        "reserved": IN,
        "shape": IN,
        "lat0": DE,
        "lon0": DE,
        "radiuslat1": DE,  # use for radius or lat1
    },
    "QTMCFGGEOFENCE_DIS": {
        "status": QS,
        "geofenceindex": IN,
        "geofencemode": IN,  # should be 0
    },
    "QTMCFGGEOFENCE_POLY": {
        "status": QS,
        "geofenceindex": IN,
        "geofencemode": IN,
        "reserved": IN,
        "shape": IN,
        "lat0": DE,
        "lon0": DE,
        "radiuslat1": DE,  # use for radius or lat1
        "lon1": DE,
        "lat2": DE,
        "lon2": DE,
        "lat3": DE,
        "lon3": DE,
    },
    "QTMCFGIMUINT": {
        "status": QS,
        "index": IN,
        "mode": IN,
        "reserved": IN,
        "actLevel": IN,
    },
    "QTMCFGIMUTC": {
        "status": QS,
        "imustatus": IN,
    },
    "QTMCFGLA": {
        "status": QS,
        "type": IN,
        "laX": DE,
        "laY": DE,
        "laZ": DE,
    },
    "QTMCFGLAM": {
        "status": QS,
        "mode": IN,
    },
    "QTMCFGMSGRATE": {
        "status": QS,
        "msgname": ST,
        "rate": IN,
        "msgver": IN,  # or offset
    },
    "QTMCFGMSGRATE_NOVER": {
        "status": QS,
        "msgname": ST,
        "rate": IN,
    },
    "QTMCFGNAVMODE": {
        "status": QS,
        "mode": IN,
    },
    "QTMCFGNMEADP": {
        "status": QS,
        "utcdp": IN,
        "posdp": IN,
        "altdp": IN,
        "dopdp": IN,
        "spddp": IN,
        "cogdp": IN,
    },
    "QTMCFGNMEATID": {
        "status": QS,
        "mainTalker": ST,
        "gsvTalker": ST,
    },
    "QTMCFGODO": {
        "status": QS,
        "state": IN,  # 0 disabled, 1 enabled
        "initdist": DE,
    },
    "QTMCFGPPS": {
        "status": QS,
        "ppsindex": IN,
        "enable": IN,
        "duration": IN,
        "ppsmode": IN,
        "polarity": IN,
        "reserved": IN,
    },
    "QTMCFGPROT": {
        "status": QS,
        "porttype": IN,
        "portid": IN,
        "inputprot": ST,  # 8 hex digits
        "outputprot": ST,  # 8 hex digits
    },
    "QTMCFGRCVRMODE": {
        "status": QS,
        "rcvrmode": IN,  # 0 unknown, 1 rover, 2 base
    },
    "QTMCFGRSID": {
        "status": QS,
        "rsid": IN,
    },
    "QTMCFGRTCM": {
        "status": QS,
        "msmtype": IN,
        "msmmode": IN,
        "msmelevthd": IN,
        "reserved1": ST,  # must be 07 with leading 0
        "reserved2": ST,  # must be 06 with leading 0
        "ephmode": IN,
        "ephinterval": IN,
    },
    "QTMCFGRTK": {
        "status": QS,
        "diffmode": IN,  # 0 disable, 1 auto, 2 RTD only
        "relmode": IN,  # 1 absolute, 2 relative
    },
    "QTMCFGRTKSRCTYPE": {
        "status": QS,
        "srctype": IN,
    },
    "QTMCFGSAT": {
        "status": QS,
        "systemid": IN,
        "signalid": ST,  # hex as string
        "masklow": ST,  # hex as string
        "maskhigh": ST,  # hex as string
    },
    "QTMCFGSAT_LOW": {
        "status": QS,
        "systemid": IN,
        "signalid": ST,  # hex as string
        "masklow": ST,  # hex as string
    },
    "QTMCFGSIGNAL": {
        "status": QS,
        "gpssig": ST,  # hex as string, default 07
        "glonasssig": ST,  # hex as string, default 03
        "galileosig": ST,  # hex as string, default 0F
        "beidousig": ST,  # hex as string, default 3F
        "qzsssig": ST,  # hex as string, default 07
        "navicsig": ST,  # hex as string, default 01
    },
    "QTMCFGSTATICHOLD": {
        "status": QS,
        "shstate": IN,
    },
    "QTMCFGSVIN": {
        "status": QS,
        "svinmode": IN,
        "cfgcnt": IN,
        "acclimit": DE,
        "ecefx": DE,
        "ecefy": DE,
        "ecefz": DE,
    },
    "QTMCFGUART": {
        "status": QS,
        "portid": IN,
        "baudrate": IN,
        "databit": IN,
        "parity": IN,
        "stopbit": IN,
        "flowctrl": IN,
    },
    "QTMCFGVEHMOT": {
        "status": QS,
        "mode": IN,
        "vehtype": IN,
    },
    "QTMDEBUGON": {"status": QS},
    "QTMDEBUGOFF": {"status": QS},
    "QTMDOP": {
        "msgver": IN,  # always 1 for this version
        "tow": IN,
        "gdop": DE,
        "pdop": DE,
        "tdop": DE,
        "vdop": DE,
        "hdop": DE,
        "ndop": DE,
        "edop": DE,
    },
    "QTMDRCAL": {
        "msgver": IN,  # always 1 for this version
        "calstate": IN,  # 0 = none, 1 = lightly, 2 = fully
        "navtype": IN,  # 0 = none, 1 = GNSS only, 2 = DR only, 3 = GNSS + DR
    },
    "QTMDRCLR": {"status": QS},
    "QTMCFGDRHOT": {"status": QS, "mode": IN},
    "QTMDRPVA": {
        "msgver": IN,  # always 1 for this version
        "timestamp": IN,
        "time": DE,
        "solType": IN,
        "lat": DE,
        "lon": DE,
        "alt": DE,
        "sep": DE,
        "velN": DE,
        "velE": DE,
        "velD": DE,
        "speed": DE,
        "roll": DE,
        "pitch": DE,
        "heading": DE,
    },
    "QTMDRSAVE": {"status": QS},
    "QTMEPE": {
        "msgver": IN,  # always 2 for this version
        "epenorth": DE,
        "epeeast": DE,
        "epedown": DE,
        "epe2d": DE,
        "epe3d": DE,
    },
    "QTMGEOFENCESTATUS": {"msgver": IN, "time": TM, "group": (4, {"staten": IN})},
    "QTMGETUTC": {
        "status": QS,
        "year": IN,
        "month": IN,
        "day": IN,
        "hour": IN,
        "minute": IN,
        "second": IN,
        "millisecond": IN,
        "reserved": ST,
        "leapsecond": IN,
    },
    "QTMGNSSSTART": {"status": QS},
    "QTMGNSSSTOP": {"status": QS},
    "QTMGPS": {
        "timestamp": IN,
        "tow": IN,
        "lat": DE,
        "lon": DE,
        "alt": DE,
        "speed": DE,
        "heading": DE,
        "accuracy": DE,
        "HDOP": DE,
        "PDOP": DE,
        "numSv": IN,
        "fixMode": IN,
    },
    "QTMJAMMINGSTATUS": {
        "msgver": IN,  # always 1 for this version
        "jammingstatus": IN,
    },
    "QTMIMU": {
        "timestamp": IN,
        "accX": DE,
        "accY": DE,
        "accZ": DE,
        "angrateX": DE,
        "angrateY": DE,
        "angrateZ": DE,
        "wheeltick": IN,
        "last_tick_timestamp": IN,
    },
    "QTMIMUTYPE": {
        "msgver": IN,  # always 1 for this version
        "imustatus": IN,  # 0 = failed, 1 = OK
    },
    "QTMINS": {
        "timestamp": IN,
        "soltype": IN,
        "lat": DE,
        "lon": DE,
        "height": DE,
        "velN": DE,
        "velE": DE,
        "velD": DE,
        "roll": DE,
        "pitch": DE,
        "yaw": DE,
    },
    "QTMLS": {
        "msgver": IN,  # always 1 for this version
        "tow": IN,
        "lsRef": HX,
        "wno": IN,
        "leapsecond": IN,
        "flag": IN,
        "lsfRef": HX,
        "reserved": ST,
        "wnlsf": IN,
        "dn": IN,
        "lsf": IN,
    },
    "QTMODO": {
        "msgver": IN,  # always 1 for this version
        "time": TM,
        "state": IN,  # 0 disabled, 1 enabled
        "dist": DE,
    },
    "QTMPL": {
        "msgver": IN,  # always 1 for this version
        "tow": IN,
        "pul": DE,
        "reserved1": IN,
        "reserved2": IN,
        "plposn": DE,
        "plpose": DE,
        "plposd": DE,
        "plveln": DE,
        "plvele": DE,
        "plveld": DE,
        "reserved3": IN,
        "reserved4": IN,
        "pltime": IN,
    },
    "QTMPVT": {
        "msgver": IN,  # always 1 for this version
        "tow": IN,
        "date": ST,  # yyyymmdd
        "time": TM,
        "reserved": IN,
        "fixtype": IN,
        "numsv": IN,
        "leaps": IN,
        "lat": DE,
        "lon": DE,
        "alt": DE,
        "sep": DE,
        "veln": DE,
        "vele": DE,
        "veld": DE,
        "spd": DE,
        "hdg": DE,
        "hdop": DE,
        "pdop": DE,
    },
    "QTMQVER": {
        "status": QS,
        "msgver": IN,
        "description": ST,
        "version": ST,
        "buildDate": ST,
        "buildTime": ST,
    },
    "QTMRESETODO": {"status": QS},
    "QTMRESTOREPAR": {"status": QS},
    "QTMSAVEPAR": {"status": QS},
    "QTMSENMSG": {
        "msgver": IN,  # 2 or 4
        "timestamp": IN,
        "group": (
            "None",
            {
                "sensor_info": DE,
            },
        ),
    },
    "QTMSN": {
        "status": QS,
        "snid": IN,  # fixed to 1
        "length": IN,
        "serialno": ST,
    },
    "QTMSTD": {
        "reserved": ST,
        "utc": TM,
        "wno": IN,
        "tow": IN,
        "stdLat": ST,  # defined as ST as fields can be blank
        "stdLon": ST,
        "stdAlt": ST,
        "stdSep": ST,
        "stdVelN": ST,
        "stdVelE": ST,
        "stdVelD": ST,
        "stdSpd": ST,
        "stdRoll": ST,
        "stdPitch": ST,
        "stdHeading": ST,
    },
    "QTMSVINSTATUS": {
        "msgver": IN,  # always 1 for this version
        "tow": IN,
        "valid": IN,
        "reserved0": IN,
        "reserved1": IN,
        "obs": IN,
        "cfgdur": IN,
        "meanx": DE,
        "meany": DE,
        "meanz": DE,
        "meanacc": DE,
    },
    "QTMTXT": {
        "msgver": IN,  # always 1 for this version
        "totalsennum": IN,
        "sennum": IN,
        "textid": IN,
        "text": ST,
    },
    "QTMUNIQID": {
        "status": QS,
        "length": IN,
        "ID": HX,
    },
    "QTMVEHATT": {
        "msgver": IN,  # always 1 for this version
        "timestamp": IN,
        "roll": DE,
        "pitch": DE,
        "heading": DE,
        "acc_Roll": DE,
        "acc_Pitch": DE,
        "acc_Heading": DE,
    },
    "QTMVEHMOT": {
        "msgver": IN,  # 1 or 2
        "group": (
            "None",
            {
                "vehicle_info": DE,
            },
        ),
    },
    "QTMVEHMSG": {
        "msgver": IN,  # 1 = speed, 2 = tick, 3 = speed 4wd, 4 = ticks 4wd
        "timestamp": IN,
        "group": (
            "None",
            {
                "vehicle_info": DE,
            },
        ),
    },
    "QTMVEL": {
        "msgver": IN,  # always 1 for this version
        "time": TM,
        "veln": DE,
        "vele": DE,
        "veld": DE,
        "gndspd": DE,
        "spd": DE,
        "hdg": DE,
        "gndspdacc": DE,
        "spdacc": DE,
        "hdgacc": DE,
    },
    "QTMVER": {
        "msgver": IN,  # always 1 for this version
        "vername": ST,
        "verstr": ST,
        "builddate": ST,  # yyyy/mm/dd
        "buildtime": ST,  # hh:mm:ss
    },
    "QTMVERNO": {
        "verstr": ST,
        "builddate": ST,  # yyyy/mm/dd
        "buildtime": ST,  # hh:mm:ss
    },
    # ***************************************************************
    # Septentrio X5 Proprietary message types
    # https://www.septentrio.com/en/products/gnss-receivers/gnss-receiver-modules/mosaic-x5#resources
    # ***************************************************************
    "SSNHRP": {
        "msgId": ST,  # "HRP"
        "utc": TM,
        "date": DT,
        "hdg": DE,  # degrees True
        "roll": DE,
        "pitch": DE,
        "hdgstd": DE,
        "rollstd": DE,
        "pitchstd": DE,
        "sip": IN,
        "attmode": CH,
        "magvar": DE,
        "magvardir": LND,
    },
    "SSNRBD": {
        "msgId": ST,  # "RBD"
        "utc": TM,
        "date": DT,
        "azi": DE,
        "ele": DE,
        "sip": IN,
        "quality": CH,
        "basemode": CH,
        "corrage": DE,
        "roverserial": ST,
        "baseid": ST,
    },
    "SSNRBP": {
        "msgId": ST,  # "RBP"
        "utc": TM,
        "date": DT,
        "baseN": DE,  # meters
        "baseE": DE,
        "baseU": DE,
        "sip": IN,
        "quality": CH,
        "basemode": CH,
        "corrage": DE,
        "roverserial": ST,
        "baseid": ST,
    },
    "SSNRBV": {
        "msgId": ST,  # "RBV"
        "utc": TM,
        "date": DT,
        "baserocN": DE,
        "baserocE": DE,
        "baserocU": DE,
        "sip": IN,
        "quality": CH,
        "basemode": CH,
        "corrage": DE,
        "roverserial": ST,
        "baseid": ST,
    },
    "SSNSNC": {
        "msgId": ST,  # "SNC"
        # "[",
        "rev": CH,
        "tim": IN,  # milliseconds
        "week": IN,
        "group": (
            "None",
            {
                # "[",
                "cdidx": ST,
                "status": QS,
                "errcode": ST,
                "info": ST,
                # "]",
            },
        ),
        # "]",
    },
    "SSNTFM": {
        "msgId": ST,  # "TFM"
        "utc": TM,
        "heightind": ST,
        "usage2122": ST,
        "usage2324": ST,
        "usage252627": ST,
    },
}
